#include "camera.h"
#include "pygmalion.h"

Camera::Camera():type(1),width(320),height(240),eyeAngle(60),zoom(0.1){
  //strcpy(name,"camera");
  sprintf(name,"camera");
  type=PERSPECTIVE;
  //eyeAngle=atan(240.0/320.0*tan(60/2*M_PI/180))/M_PI*180*2.0;
  eyeAngle=POVAng2OGLAng(*this,60);
}

Camera::~Camera(){
}

Camera::Camera(const Camera& rhs){
  type=rhs.type;
  strcpy(name,rhs.name);
  view=rhs.view;
  camera=rhs.camera;
  width=rhs.width;
  height=rhs.height;
  eyeAngle=rhs.eyeAngle;
  zoom=rhs.zoom;
}

const Camera&
Camera::operator=(const Camera& rhs){
  if (this==&rhs)
    return *this;
  
  type=rhs.type;
  strcpy(name,rhs.name);
  view=rhs.view;
  camera=rhs.camera;
  width=rhs.width;
  height=rhs.height;
  eyeAngle=rhs.eyeAngle;
  zoom=rhs.zoom;
  return *this;  
}

void
Camera::WriteFile(FILE* fp){
  switch(type){
  case PERSPECTIVE:
    fprintf(fp,"perspective\n");
    break;
  case XYMP:
    fprintf(fp,"xymp\n");
    break;
  case XYPM:
    fprintf(fp,"xypm\n");
    break;
  case YZMP:
    fprintf(fp,"yzmp\n");
    break;
  case YZPM:
    fprintf(fp,"yzpm\n");
    break;
  case XZMP:
    fprintf(fp,"xzmp\n");
    break;
  case XZPM:
    fprintf(fp,"xzpm\n");
    break;
  }

  if(strlen(name)!=0){
    fprintf(fp,"%s\n",name);
  }else{
    fprintf(fp,"Untitled\n");
  }

  switch(type){
  case PERSPECTIVE:
    view.WriteFile(fp);
    break;
  case XYMP:
  case XYPM:
  case YZMP:
  case YZPM:
  case XZMP:
  case XZPM:
    fprintf(fp,"%f\n",zoom);
    break;
  }
  camera.WriteFile(fp);
  fprintf(fp,"%d %d %f\n",width,height,eyeAngle);
}

void
Camera::ReadFile(FILE* fp){
  char buf[255];
  fscanf(fp,"%s ",buf); // type
  for(int cnt1=0;cnt1<255;cnt1++){
    name[cnt1]=0;
  }
  // perspective
  if(!strcmp(buf,"normal")){ // back compati
    type=PERSPECTIVE;
  }
  if(!strcmp(buf,"perspective")){
    type=PERSPECTIVE;
  }
  // XY
  if(!strcmp(buf,"xy")){ // back compati
    type=XYPM;
  }
  if(!strcmp(buf,"xypm")){ 
    type=XYPM;
  }
  if(!strcmp(buf,"xymp")){ 
    type=XYMP;
  }
  // YZ
  if(!strcmp(buf,"yz")){ // back compati
    type=YZPM;
  }
  if(!strcmp(buf,"yzpm")){
    type=YZPM;
  }
  if(!strcmp(buf,"yzmp")){
    type=YZMP;
  }
  // XZ
  if(!strcmp(buf,"xz")){ // back compati
    type=XZPM;
  }
  if(!strcmp(buf,"xzpm")){
    type=XZPM;
  }
  if(!strcmp(buf,"xzmp")){
    type=XZMP;
  }
  
  
  fgets(buf,254,fp); // name
#ifdef DEBUG
  cout << "buf==" << buf << endl;
  cout << "Name has endline?" << endl;
#endif
  strncpy(name,buf,strlen(buf)-1);
  name[strlen(name)]='\0';
#ifdef DEBUG
  cout << "name==" << name << "  ";
  cout << "desu." << endl;
#endif
    
  switch(type){
  case PERSPECTIVE:
    fscanf(fp,"%lf %lf %lf",&view.coord[0],&view.coord[1],&view.coord[2]);
#ifdef DEBUG
    cout << "view==";
    view.Print();
#endif
    break;
  case XYMP:
  case XYPM:
  case YZMP:
  case YZPM:
  case XZMP:
  case XZPM:
    fscanf(fp,"%f ",&zoom);
    break;
  }
  fscanf(fp,"%lf %lf %lf",&camera.coord[0],
	 &camera.coord[1],&camera.coord[2]);
  fscanf(fp,"%d %d %f",&width,&height,&eyeAngle);
#ifdef DEBUG
  cout << "camera==";
  camera.Print();
  cout << "width,height,angle == " << width << ", " 
       << height << ", " << eyeAngle << endl;
#endif
}
