#include "edge.h"
#include "face.h"
#include <cstdio>
#include <FL/gl.h>

using namespace std;
extern int vertexOffset;
extern int faceOffset;
 
// Constructor 1
Edge::Edge():edgePointLink(-1){
}

// Constructor 2
// vNum1_ < vNum2_
Edge::Edge(const int vNum1_,const int vNum2_) :
  edgePointLink(-1){
  vertexLink.push_back(vNum1_); // Left vertex
  vertexLink.push_back(vNum2_); // Right vertex
  weight=(vertex[vNum1_].weight+vertex[vNum2_].weight)/2;
}

// Copy Constructor
Edge::Edge(const Edge& rhs):edgePointLink(rhs.edgePointLink),
    weight(rhs.weight){  
  facePLink.erase(facePLink.begin(),facePLink.end());
  vertexLink.erase(vertexLink.begin(),vertexLink.end());
  for(int cnt1=0;cnt1<rhs.facePLink.size();cnt1++){
    facePLink.push_back(rhs.facePLink[cnt1]);
  }
  for(int cnt1=0;cnt1<rhs.vertexLink.size();cnt1++){
    vertexLink.push_back(rhs.vertexLink[cnt1]);
  }  
}

// Destructor
Edge::~Edge(){
}

// Copy Constructor
const Edge&
Edge::operator=(const Edge& rhs){
  if (this==&rhs)
  return *this;

  //vertexLink=rhs.vertexLink;
  //facePLink=rhs.facePLink;
  facePLink.erase(facePLink.begin(),facePLink.end());
  vertexLink.erase(vertexLink.begin(),vertexLink.end());
  for(int cnt1=0;cnt1<rhs.facePLink.size();cnt1++){
    facePLink.push_back(rhs.facePLink[cnt1]);
  }
  for(int cnt1=0;cnt1<rhs.vertexLink.size();cnt1++){
    vertexLink.push_back(rhs.vertexLink[cnt1]);
  }  
  edgePointLink=rhs.edgePointLink;
  weight=rhs.weight;
  return *this;
}

// Why I made this procedure?
// Return Vertex
const Vertex&
Edge::ReturnVertex(const int LR){
  return vertex[vertexLink[LR]];
}

// Calculate Edge Point
Vertex
Edge::CalcEdgePoint(){
  Vertex tmp1,tmp2;
  //Vertex tmp3; // experimental use
  int facePLinkSize;

  tmp1.Plus(vertex[vertexLink[Right]]);
  tmp1.Plus(vertex[vertexLink[Left]]);
  
  /*
  // experimental
  tmp3=tmp1;
  tmp3.Multi(0.5);
  */

  facePLinkSize=facePLink.size();
  for(int cnt1=0;cnt1<facePLinkSize;cnt1++){
    tmp2.Plus(face[facePLink[cnt1]].ReturnFacePoint());
  }
  if(facePLinkSize==0){
    tmp1.Multi(0.5);
  }else {
    tmp1.Multi(weight);
    tmp2.Multi(1.0/facePLinkSize);
    tmp2.Multi(2.0*(1.0-weight));
    tmp1.Plus(tmp2);
    tmp1.Multi(0.5);
  }
  tmp1.weight=weight;
  return tmp1; 
  
  /*
  // experimental
  // ADD
  tmp1.Multi(1.0-weight);
  tmp3.Multi(weight);
  tmp1.Plus(tmp3);
  tmp1.weight=weight;
  return tmp1;
  */

}

const Vertex&
Edge::ReturnEdgePoint(){
  return vertex[edgePointLink];
}

void
Edge::ConnectToFacePoint(){
  int facePLinkSize=facePLink.size();
  for(int cnt1=0;cnt1<facePLinkSize;cnt1++){
    /*
    cout << "facePLink[" << cnt1 << "]==" << facePLink[cnt1] << endl;
#ifdef DEBUG
    if(facePLink[cnt1]>=face.size()){
      cout << "Strange!!!!! facePLink[" << cnt1 << "]==" 
	   << facePLink[cnt1] << ">=face.size()" << endl;
    }
#endif
    */
    AddEdge(face[facePLink[cnt1]].facePointLink,edgePointLink);
  }
}

/*
// Get Middle Point of 2 Vertex Points
Vector3d
Edge::GetMid(){
  Vector3d mid;
  mid=vertex[vertexLink[Right]]+vertex[vertexLink[Left]];
  mid.Multi(0.5);
  return mid;
}
*/

// Print
void
Edge::Print(){
  cout << "vertex number : "<< vertexLink[Left] << " " << vertexLink[Right] << endl;
}

void
Edge::Info(){
  Print();
  for(int cnt1=0;cnt1<facePLink.size();cnt1++){
    cout << "facePLink[" << cnt1 << "]==" << facePLink[cnt1] << endl;
  }
  cout << "weight:" << weight << endl;
}

// Display Edge on GL Window
void
Edge::Draw(){
  glBegin(GL_LINE_STRIP);
  vertex[vertexLink[Left]].Draw();
  vertex[vertexLink[Right]].Draw();
  glEnd();
}

// Set Face Link
void
Edge::SetFaceLink(){
  facePLink.push_back(face.size()-1);
}

void
Edge::UnSetFaceLink(int num){
#ifdef DEBUG
  cout << endl << "UnSetFaceLink Face[" << num << "]" << " in Edge" << endl;
  cout << "facePLink" << endl;
  for(int cnt1=0;cnt1<facePLink.size();cnt1++){
    cout << facePLink[cnt1] << " ";
  }
  cout << endl;
#endif
  //vector<int>::iterator end=remove(facePLink.begin(),facePLink.end(),num);  
  //facePLink.erase(end,facePLink.end());
  for(int cnt1=0;cnt1<facePLink.size();cnt1++){
    if(facePLink[cnt1]==num){
      facePLink.erase(&facePLink[cnt1]);
      return;
    }
  }
  
#ifdef DEBUG
  cout << "UnSetFaceLink Face[" << num << "]" << endl;
  cout << "facePLink" << endl;
  for(int cnt1=0;cnt1<facePLink.size();cnt1++){
    cout << facePLink[cnt1] << " ";
  }
  cout << endl;
#endif
}

/*
// Find Face and substitute Edge orders
// WARNING !
// Edge::vertex and vector<Vertex>& vertex_ exist!
int
Edge::FindAddFace(){
  Vertex* v1=&(vertex[vertexLink[Left]]);
  Vertex* v2=&(vertex[vertexLink[Right]]);
  int awayV1; // Index of away Vertex of edge that is adjucent this edge
  int awayV2;
  int edge1,edge2,flag=0; // Three Face generates, flag=1

  // v1
  for(int cnt1=0;cnt1<v1->edgePLink.size();cnt1++){
    if(this==&(edge[v1->edgePLink[cnt1]])){
      continue;
    }
    awayV1=edge[v1->edgePLink[cnt1]].vertexLink[Left];
    if(&(vertex[awayV1])==v1){
      awayV1=edge[v1->edgePLink[cnt1]].vertexLink[Right];
    }
    
    // v2
    for(int cnt2=0;cnt2<v2->edgePLink.size();cnt2++){
      if(this==&(edge[v2->edgePLink[cnt2]])){
	continue;
      }
      awayV2=edge[v2->edgePLink[cnt2]].vertexLink[Left];
      if(&(vertex[awayV2])==v2){
	awayV2=edge[v2->edgePLink[cnt2]].vertexLink[Right];
      }
      // Three Edges Face
      if(awayV1==awayV2){
#ifdef DEBUG
	cout << "Found connectable three edges!" << endl;
#endif
	edge1=v1->edgePLink[cnt1];
	edge2=v2->edgePLink[cnt2];
	AddFace((int)(edge.size()-1),edge1,edge2,
		vertexLink[Left],vertexLink[Right],awayV1);
	flag=1;
      }

      // Four Edges Face
      int flag2,swapV1,swapV2;
      swapV1=awayV1;
      swapV2=awayV2;
      if(swapV1>swapV2){
	swap(swapV1,swapV2);
      }
      flag2=FindEdge(swapV1,swapV2);
      if(flag2!=-1){
#ifdef DEBUG
	cout << "Found connectable four Edges Face!" << endl;
#endif
	AddFace((int)edge.size()-1,flag2,v1->edgePLink[cnt1],v2->edgePLink[cnt2],
		vertexLink[Left],awayV1,awayV2,vertexLink[Right]);
      }
    }
  }
  return flag;
}
*/

// Find Face and substitute Edge orders
// WARNING !
// Edge::vertex and vector<Vertex>& vertex_ exist!
int
Edge::FindAddFace(){
  Vertex* v1=&(vertex[vertexLink[Left]]);
  Vertex* v2=&(vertex[vertexLink[Right]]);
  int awayV1; // Index of away Vertex of edge that is adjucent this edge
  int awayV2;
  int edge1,edge2,flag=0; // Three Face generates, flag=1
  int edgePLinkSize1=v1->edgePLink.size();
  int edgePLinkSize2=v2->edgePLink.size();

  // v1    
  for(int cnt1=0;cnt1<edgePLinkSize1;cnt1++){
    if(this==&(edge[v1->edgePLink[cnt1]])){
      continue;
    }
    awayV1=edge[v1->edgePLink[cnt1]].vertexLink[Left];
    if(&(vertex[awayV1])==v1){
      awayV1=edge[v1->edgePLink[cnt1]].vertexLink[Right];
    }
    
    // v2    
    for(int cnt2=0;cnt2<edgePLinkSize2;cnt2++){
      if(this==&(edge[v2->edgePLink[cnt2]])){
	continue;
      }
      awayV2=edge[v2->edgePLink[cnt2]].vertexLink[Left];
      if(&(vertex[awayV2])==v2){
	awayV2=edge[v2->edgePLink[cnt2]].vertexLink[Right];
      }

      if(nowStatus==sNORMAL){
	// Three Edges Face
	if(awayV1==awayV2){
#ifdef DEBUG
	  cout << "Found connectable three edges!" << endl;
#endif
	  edge1=v1->edgePLink[cnt1];
	  edge2=v2->edgePLink[cnt2];
	  AddFace((int)(edge.size()-1),edge1,edge2,
		  vertexLink[Left],vertexLink[Right],awayV1);
	  flag=1;
	}
      }
      
      // Four Edges Face
      int awayV3,flag2;//,swapV1,swapV2;
      /*
      swapV1=awayV1;
      swapV2=awayV2;
      if(swapV1>swapV2){
	swap(swapV1,swapV2);
      }
      flag2=FindEdge(swapV1,swapV2);
      if(flag2!=-1){
      */
      for(int cnt3=0;cnt3<vertex[awayV1].edgePLink.size();cnt3++){
	awayV3=edge[vertex[awayV1].edgePLink[cnt3]].vertexLink[Left];
	if(awayV1==awayV3){
	  awayV3=edge[vertex[awayV1].edgePLink[cnt3]].vertexLink[Right];
	}
	if(awayV3==awayV2){
#ifdef DEBUG
	  cout << "Found connectable four Edges Face!" << endl;
#endif
	  AddFace((int)edge.size()-1,vertex[awayV1].edgePLink[cnt3],
		  v1->edgePLink[cnt1],v2->edgePLink[cnt2],
		  vertexLink[Left],awayV1,awayV2,vertexLink[Right]);
	}
      }
    }
  }
  return flag;
}

void
Edge::MoreSharp(){
  weight+=0.1;
  if(weight>1.0){
    weight=1.0;
  }
}

void
Edge::MoreSmooth(){
  weight-=0.1;
  if(weight<0){
    weight=0;
  }
}

void
Edge::WriteFile(FILE* fp){
  fprintf(fp,"%d %d %d %f\n",vertexLink[Left],vertexLink[Right],
	  facePLink.size(),weight);
  // facePLink[]
  for(int cnt1=0;cnt1<facePLink.size();cnt1++){
    fprintf(fp,"%d ",facePLink[cnt1]);
  }
  fprintf(fp,"\n");
}

void
Edge::ReadFile(FILE* fp){
  int tmpFacePLinkSize,leftV,rightV,tmpFace;
  
  fscanf(fp,"%d %d %d %f",&leftV,&rightV,&tmpFacePLinkSize,&weight);
  vertexLink.push_back(leftV+vertexOffset);
  vertexLink.push_back(rightV+vertexOffset);
  for(int cnt1=0;cnt1<tmpFacePLinkSize;cnt1++){
    fscanf(fp,"%d",&tmpFace);
    facePLink.push_back(tmpFace+faceOffset);
  }
  edgePointLink=-1;
}

int
Edge::AnotherVertex(int one){
  if(one==vertexLink[0]){
    return vertexLink[1];
  }else{
    return vertexLink[0];
  }
}

// This procedure should be called in edge subjecting to two faces !!
int
Edge::AnotherFace(int one){
  if(one==facePLink[0]){
    return facePLink[1];
  }else{
    return facePLink[0];
  }
}
