#include <algorithm>
#include <FL/math.h>

#include <FL/Fl.H>                      // Main FLTK header file
#include <FL/Fl_Group.H>                // Fl_Group header file
#include <FL/Fl_Window.H>               // Fl_Window header file
#include <FL/fl_ask.H>                  // FLTK convenience functions
#include <FL/fl_file_chooser.H>         // FLTK file chooser
#include <FL/Fl_Menu_Bar.H>       
#include <FL/Fl_Button.H>
#include <FL/fl_draw.H>

#include "vector3d.h"
#include "vertex.h"
#include "edge.h"
#include "face.h"
#include "glWindow.h"
#include "glWindowUI.h"
#include "pygmalion.h"
#include "light.h"
#include "cameraset.h" 
#include "mainmenu.h"


using namespace std;

const int BufSize=512; // Buffer Size in picking
//const int PICKSIZE=5; // test value
const int EFFECT=4; // pivot,approach,zoom

const int ExtraR=3; // extra radius in Ortho window when VertexPick()

const int SelectTimes=4; // this decide vagueness in 

const float PointSize=3.0; // drawing size of non-selected points in Display()

double mostDepth; // most depth value
GLuint selectBuf[BufSize]; // common for Pick

vector<Camera> cameraList;
vector<Light> lightList;

int xx=0,yy=0; // Mouse trans

GLfloat light0_pos[] = {1.0, 3.0, 1.0, 0}; // position
GLfloat light0_dif[] = {1.0, 1.0, 1.0, 1.0}; // diffuse
GLfloat light1_pos[] = {-1.0, -3.0, -1.0, 0}; // position
GLfloat light1_dif[] = {1.0, 1.0, 1.0, 1.0}; // diffuse
GLfloat material_dif0[] = {0.9, 0.9, 1.0, 1.0}; // diffuse
GLfloat material_amb0[] = {0.0, 0.0, 0.0, 1.0}; // ambient

extern vector<int> selectV;
extern vector<int> selectE;
extern vector<int> selectF;
extern vector<int> totalV;
int nowV; // now selected Vertex
Vector3d newV; // temporary generated Vertex whlie Push()ing or Drag()ing
GLboolean newFlag; // flag Push()ing or Drag()ing
int nowE;
int nowF;

// {min,max}{X,Y,Z} are real {min,max} and for right(OpenGL) coordinate 
// For left(POV-Ray) coordinate, should change Z
double minX=-0.1,minY=-0.1,minZ=-0.1; // clipping
double maxX=0.1,maxY=0.1,maxZ=0.1; // clipping
int minXV,minYV,minZV,maxXV,maxYV,maxZV; //{min,max}{X,Y,Z} vertex

GLdouble z1; // for calculation a plane that on selected vertex
Vector3d start,end; // for calculation vertex movement

extern vector<Vertex> vertex; // Vertex Vector
extern vector<Edge> edge; // Edge Vector
extern vector<Face> face; // Face Vector

extern int subdivideLevel;
extern vector<int> subdividedVertex; // Vertex
extern vector<int> subdividedEdge; // Edge
extern vector<int> subdividedFace; // Face

extern MainMenu* mainM;
extern vector<GLWindowUI*> glW;
extern char* filename;

extern double minXS,maxXS,minYS,maxYS,minZS,maxZS;

extern Fl_Menu_Item* restX; // restrict X
extern Fl_Menu_Item* restY;
extern Fl_Menu_Item* restZ;

enum {
  sREAL,sRELEASE
}syncDraw=sREAL; // real time syncing {real time,mouse release}

const int mConnect=1;
const int mTrans=2;
const int mScale=3;
const int mRotate=4;
int lMode=mConnect; // left button mode

enum {
  sVertex,sEdge,sFace
}sMode=sVertex; // select mode

enum {
  mL,mM,mR // Left Middle Right
}motBut; // Motion Button

extern enum {
  uGEOMETRY=1,
  uSELECTION=2,
  uCAMERA=4,
  uTEXTURE=8
}opCode;

// Vertex moving constraint
GLboolean xConst = GL_FALSE; 
GLboolean yConst = GL_FALSE;
GLboolean zConst = GL_FALSE;

// syncronize all GLWindow
void
Sync(){
  if(syncDraw==sREAL){
    for(int cnt1=0;cnt1<glW.size();cnt1++){
      if(glW[cnt1]->mainWindow->visible()){
	glW[cnt1]->glWindow->redraw();
      }
    }
  }  
}

Vector3d
PassP(){
  if(selectV.size()){ // return last element
    return vertex[selectV.back()];
  }else{
    Vector3d tmp;
    return tmp;
  }
}

void
CheckBB(Vertex& poi){
  if(poi.X()>maxX){
    maxX=poi.X();
    DebugOut("expand in x");
  }else if(poi.X()<minX){
    minX=poi.X();
  }
  if(poi.Y()>maxY){
    maxY=poi.Y();
  }else if(poi.Y()<minY){
    minY=poi.Y();
  }
  if(poi.Z()>maxZ){
    maxZ=poi.Z();
  }else if(poi.Z()<minZ){
    minZ=poi.Z();
  }
}


void
SetConnect(){
  lMode=mConnect;
  SelectVertex();
  mainM->connectB->setonly();
  //cout << "lMode connect" << endl;
  //cout << "select mode : Vertex" << endl;
}

void
SetRotate(){
  lMode=mRotate;
  mainM->rotateB->setonly();
  //cout << "mode rotation" << endl;
}

void
SetScale(){
  lMode=mScale;
  mainM->scaleB->setonly();
  //cout << "mode scaling" << endl;
}

void
SetTrans(){
  lMode=mTrans;
  mainM->transB->setonly();
  //cout << "mode translation" << endl;
}

void
SelectVertex(){
  sMode=sVertex;
  mainM->svertexB->setonly();
}

void
SelectEdge(){
  sMode=sEdge;
  mainM->sedgeB->setonly();
}

void
SelectFace(){
  sMode=sFace;
  mainM->sfaceB->setonly();
}


//GLWindow::GLWindow(int X,int Y,int W,int H,const char* L,int _camNum):
GLWindow::GLWindow(int X,int Y,int W,int H,const char* L):
  Fl_Gl_Window(X, Y, W, H, L),axeFlag(GL_TRUE),
  vFlag(GL_TRUE),eFlag(GL_TRUE),fFlag(GL_TRUE),
  sVFlag(GL_FALSE),sEFlag(GL_FALSE),sFFlag(GL_TRUE){
}

GLWindow::~GLWindow(){
}

/*
// copy constructor
const GLWindow&
GLWindow::operator=(const GLWindow&){
}
*/

void
GLWindow::CoordToAngle(Vector3d& tmp,float& tmpAngleY,float& tmpAngleXZ){
  double tmpX,tmpY,tmpZ,preX;
  // under is lack of my power
  tmpX=tmp.Z(); 
  tmpY=tmp.Y();
  tmpZ=tmp.X();
  tmpAngleY=asin(tmpY/cameraR)/M_PI*180;
  preX=cameraR*cos(M_PI/180*tmpAngleY);
  tmpAngleXZ=asin(tmpX/preX)/M_PI*180;
  if(tmp.X()<0){
    tmpAngleXZ=180-tmpAngleXZ;
  }
#ifdef DEBUG
  cout << "tmpAngleY,tmpAngleXZ==" << tmpAngleY << ", " << tmpAngleXZ << endl;
#endif
}
  
void
GLWindow::SetCamera(int _camNum){
  Vector3d tmp;
  camNum=_camNum;
  camera=&cameraList[camNum];  
  tmp=camera->camera;
  tmp.Minus(camera->view);
  cameraR=sqrt(tmp.Dot(tmp));

  //if(camera->type==NORMAL){
  if(camera->type==PERSPECTIVE){
    CoordToAngle(tmp,angleY,angleXZ);
#ifdef DEBUG
    cout << "angleY,angleXZ==" << angleY << "," << angleXZ << endl;
#endif
  }
  // window size
#ifdef DEBUG
  cout << "in SetCamera,GLWindow:w(),h()==" << w() << ", " << h() << endl;
#endif
  size(camera->width,camera->height);
#ifdef DEBUG
  cout << "in SetCamera,GLWindow:w(),h()==" << w() << ", " << h() << endl;
#endif
  SetCameraType();

  redraw();
}

void
GLWindow::SetCameraType(){
  if(camera->type==PERSPECTIVE){
    //MiddleDrag=&GLWindow::Pan;    
    //RightDrag=&GLWindow::Pivot;    
    glW[camNum]->SetPan(1); // 1:middle
    glW[camNum]->SetPivot(0); // 0:right
    ShiftRightDrag=&GLWindow::Approach;
    AltRightDrag=&GLWindow::Pan;
  }else{
    //MiddleDrag=&GLWindow::Pan;    
    //RightDrag=&GLWindow::Zoom;
    glW[camNum]->SetPan(1);
    glW[camNum]->SetZoom(0);
    AltRightDrag=&GLWindow::Pan;
    ShiftRightDrag=&GLWindow::Pan;
  }
}

void
GLWindow::draw(){
#ifdef DEBUG
  cout << "in draw(): glWin[" << camNum << "]" << endl;
#endif
  if(!valid()){
#ifdef DEBUG
    cout << "!valid()" << endl;
#endif
    Init();
  }
  Reshape();

  Display();
}

void
GLWindow::PreConnect(){
  int x=Fl::event_x();
  int y=Fl::event_y();

  PreSub();
  // add
  if (!valid()) {
    Init();
    Reshape();
    valid(1); // stop it from doing this next time
  }	
  newFlag=GL_TRUE;
  ConnectMode(x,y,SelectTimes,GL_FALSE);
  PostSub();
}

void
GLWindow::Push(){
  int x=Fl::event_x();
  int y=Fl::event_y();
  xx=x;
  yy=y;
  
  mostDepth=DBL_MAX;

  switch(Fl::event_key()){
  case FL_Button+1: // Left Button ///////////////////////
    if(Fl::event_state(FL_CTRL)){ // if CTRL
      break;
    }
    switch(lMode){
    case mConnect:
      PreConnect();
      break;
    }
    Sync();
    break;

  case FL_Button+2: // Middle Button ///////////////////////
    motBut=mM;
    xx=x;
    yy=y;
    break;
    
  case FL_Button+3: // Right Button ////////////////////////
    motBut=mR;
    xx=x;
    yy=y;
    break;
  }
}

void
GLWindow::Release(){
  int x=Fl::event_x();
  int y=Fl::event_y();
  mostDepth=DBL_MAX;

  if(Fl::event_state(FL_CTRL)&& Fl::event_key()==FL_Button+1){
    DebugOut("FL_RELEASE with CTRL");
    // add
    if (!valid()) {
      Init();
      Reshape();
      valid(1); // stop it from doing this next time
    }	
    SelectMode(x,y,11); // 11 means > selection radius in window.
    
    Sync();
    PushOP(uSELECTION);
    return;
  }

  // Left Button /////////////////////////
  if(Fl::event_key()==FL_Button+1){
    newFlag=GL_FALSE;
    fl_cursor(FL_CURSOR_DEFAULT,FL_WHITE,FL_BLACK);	      
    motBut=mL;
    xx=x;
    yy=y;
    if(!Fl::event_state(FL_CTRL)){
      DebugOut("LEFT_DOWN without CTRL");
      switch(lMode){
      case mConnect:
	PreSub();
	// add
	if (!valid()) {
	  Init();
	  Reshape();
	  valid(1); // stop it from doing this next time
	}	
	ConnectMode(x,y,SelectTimes,GL_TRUE);
	PostSub();
	break;
      }
      // needed mTrans...
      Sync();
      PushOP(uGEOMETRY|uSELECTION);
    }
  }
}


void
GLWindow::Drag(){
  int x=Fl::event_x();
  int y=Fl::event_y();
  Vector3d movement,tmp1,tmp2;
  Vector3d G; // average of selected Vertices
  double matX[9],matY[9]; // [9] mean [3][3]
  // for ortho
  GLdouble objX1,objY1,objZ1;
  GLdouble objX2,objY2,objZ2;

  if(!Fl::event_state(FL_CTRL)){
    if(Fl::event_state(FL_BUTTON1)){
      float oldDist,nowDist,tmpMul; // for scale
      switch(lMode){
      case mConnect: // Connect ////////////////////////////
	PreConnect();
	break;
      case mTrans: // Trans ///////////////////////////////////
	switch(camera->type){
	case PERSPECTIVE:
	  movement=PointOnPlane(vector1,PassP(),x,y);
	  tmp1=PointOnPlane(vector1,PassP(),xx,yy);
	  movement.Minus(tmp1);
	  break;
	default:
	  gluUnProject(x,camera->height-y,1.0, 
		       modelMatrix,projMatrix,viewport,
		       &objX1,&objY1,&objZ1);
	  gluUnProject(xx,camera->height-yy,1.0, 
		       modelMatrix,projMatrix,viewport,
		       &objX2,&objY2,&objZ2);
	  movement.Set(objX1-objX2,objY1-objY2,objZ1-objZ2);
	  break;
	}
	if(xConst==GL_TRUE){
	  movement.coord[0]=0;
	}
	if(yConst==GL_TRUE){
	  movement.coord[1]=0;
	}
	if(zConst==GL_TRUE){
	  movement.coord[2]=0;
	}
	for(int cnt1=0;cnt1<totalV.size();cnt1++){
	  vertex[totalV[cnt1]].Plus(movement);
	  // check max,min
	  CheckBB(vertex[totalV[cnt1]]);
	}
	for(int cnt1=0;cnt1<subdivideLevel;cnt1++){
	  RecalcSub(cnt1);
	}
	break;
      
      case mScale: // Scale /////////////////////////////////////
	for(int cnt1=0;cnt1<totalV.size();cnt1++){
	  G.Plus(vertex[totalV[cnt1]]);
	}
	G.Multi(1.0/totalV.size()); // G:average of total selected  Vertices
	oldDist=sqrt((xx-viewport[2]/2.0)*(xx-viewport[2]/2.0)+
		     (yy-viewport[3]/2.0)*(yy-viewport[3]/2.0));
	nowDist=sqrt((x-viewport[2]/2.0)*(x-viewport[2]/2.0)+
		     (y-viewport[3]/2.0)*(y-viewport[3]/2.0));  
	//tmpMul=nowDist/oldDist;
	tmpMul=(nowDist-oldDist)/100;
	DebugOut("oldDist,nowDist,tmpMul=" << oldDist << ", " 
		 << nowDist << ", " << tmpMul);
	for(int cnt1=0;cnt1<totalV.size();cnt1++){
	  movement.Set(0,0,0);
	  movement.Plus(vertex[totalV[cnt1]]);
	  movement.Minus(G);
	  //movement.Multi((double)(xx-x+yy-y)/5000.0*cameraR);
	  movement.Multi(tmpMul); // TODO why bad?
	  // constraint
	  if(xConst==GL_TRUE){
	    movement.coord[0]=0;
	  }
	  if(yConst==GL_TRUE){
	    movement.coord[1]=0;
	  }
	  if(zConst==GL_TRUE){
	    movement.coord[2]=0;
	  }
	  vertex[totalV[cnt1]].Plus(movement);
	  // check max,min
	  CheckBB(vertex[totalV[cnt1]]);
	}
	for(int cnt1=0;cnt1<subdivideLevel;cnt1++){
	  RecalcSub(cnt1);
	}      
	break;
      
      case mRotate: // Rotate //////////////////////////////////
	for(int cnt1=0;cnt1<totalV.size();cnt1++){
	  G.Plus(vertex[totalV[cnt1]]);
	}
	G.Multi(1.0/totalV.size()); // G:average of total selected Vertices
	Vector3d orthoTmp;
	switch(camera->type){
	  //case NORMAL:
	case PERSPECTIVE:
	  // for perspective
	  CalcMat33(matX,nowUp,(double)x-xx);
	  tmp2=vector1.Cross(nowUp);
	  CalcMat33(matY,tmp2,(double)y-yy);
	  if(abs(x-xx)>abs(y-yy)){
	    Identity(matY);
	  }else {
	    Identity(matX);
	  }
	  break;
	  //case XY:
	case XYPM:
	  orthoTmp.Set(0,0,1);
	  CalcMat33(matX,orthoTmp,(float)x-xx+y-yy);
	  Identity(matY);	
	  break;
	case XYMP:
	  orthoTmp.Set(0,0,-1);
	  CalcMat33(matX,orthoTmp,(float)x-xx+y-yy);
	  Identity(matY);	
	  break;
	  //case YZ:
	case YZPM:
	  orthoTmp.Set(1,0,0);
	  CalcMat33(matX,orthoTmp,(float)x-xx+y-yy);
	  Identity(matY);	
	  break;
	case YZMP:
	  orthoTmp.Set(-1,0,0);
	  CalcMat33(matX,orthoTmp,(float)x-xx+y-yy);
	  Identity(matY);	
	  break;
	  //case XZ:
	case XZPM:
	  orthoTmp.Set(0,1,0);
	  CalcMat33(matX,orthoTmp,(float)x-xx+y-yy);
	  Identity(matY);	
	  break;
	case XZMP:
	  orthoTmp.Set(0,-1,0);
	  CalcMat33(matX,orthoTmp,(float)x-xx+y-yy);
	  Identity(matY);	
	  break;
	}
	for(int cnt1=0;cnt1<totalV.size();cnt1++){
	  tmp1=vertex[totalV[cnt1]];	
	  vertex[totalV[cnt1]].Minus(G);
	  vertex[totalV[cnt1]].MultiMat33(matX);
	  vertex[totalV[cnt1]].MultiMat33(matY);
	  vertex[totalV[cnt1]].Plus(G);
	  // check max,min
	  CheckBB(vertex[totalV[cnt1]]);
	}
	for(int cnt1=0;cnt1<subdivideLevel;cnt1++){
	  RecalcSub(cnt1);
	}
	break;
      }
      Sync(); 
      SetChanged(1);    
    }else{
      if(Fl::event_state(FL_BUTTON3) && Fl::event_state(FL_SHIFT)){
	(this->*(this->ShiftRightDrag))();
      }else{
	if(Fl::event_state(FL_BUTTON3) && Fl::event_state(FL_ALT)){
	  (this->*(this->AltRightDrag))();	
	}else{
	  if(Fl::event_state(FL_BUTTON2)){
	    (this->*(this->MiddleDrag))();
	  }else{    
	    if(Fl::event_state(FL_BUTTON3)){
	      (this->*(this->RightDrag))();
	    }
	  }
	}    
      }
      redraw();
    }   
  }
  
  xx=Fl::event_x();
  yy=Fl::event_y();
}

int
GLWindow::handle(int event) {
    switch(event) {
      case FL_PUSH:
	make_current(); // make OpenGL context current
	Push();
	//... mouse down event ...
	//... position in Fl::event_x() and Fl::event_y()
	return 1;
      case FL_DRAG:
	make_current(); // make OpenGL context current
	Drag();
	//... mouse moved while down event ...
	return 1;
      case FL_RELEASE:    
	make_current(); // make OpenGL context current	
	Release();
	//... mouse up event ...
	return 1;
      case FL_FOCUS :
      case FL_UNFOCUS :
	//... Return 1 if you want keyboard events, 0 otherwise
	return 1;
      case FL_KEYBOARD:
	return Key();
      case FL_SHORTCUT:
	//... shortcut, key is in Fl::event_key(), ascii in Fl::event_text()
	//... Return 1 if you understand/use the shortcut event, 0 otherwise...
	//return 1;
	return 0; // "return 1" prevent glWindow gets glW's event
    case FL_HIDE:
      DebugOut ("GLWindow closed" << endl);
      mainM->HideViewIndex(camNum);
      return 1;
    case FL_SHOW:
      DebugOut ("GLWindow showed"<< endl);
      mainM->ShowViewIndex(camNum);
      return 1;
    case FL_ENTER:
      //glW[camNum]->glWindow->take_focus();
      //glW[camNum]->glWindow->set_non_modal();
      glW[camNum]->mainWindow->show();
      DebugOut ("enter glW[" << camNum << "]" << endl);
      return 1;
    default:
	// tell FLTK that I don't understand other events
	return 0;
    }
}


void
GLWindow::CalcCameraPos(){
  Vector3d tmpVec1;
  double tmpX,tmpY,tmpZ;
  
  tmpY=cameraR*sin(M_PI/180*angleY);
  tmpX=cameraR*cos(M_PI/180*angleY);
  tmpZ=tmpX*sin(M_PI/180*angleXZ);
  tmpX*=cos(M_PI/180*angleXZ);
  tmpVec1.Set(tmpX,tmpY,tmpZ);  
  tmpVec1.Plus(camera->view);
  camera->camera=tmpVec1;
#ifdef DEBUG
  cout << "angleXZ,angleY==" << angleXZ << ", " << angleY << endl;
  cout << "camera==";
  camera->camera.Print();
  cout << "view==";
  camera->view.Print();
#endif
}

// for perspective window
void
//GLWindow::Pivot(int x,int y){
GLWindow::Pivot(){
  int x=xx-Fl::event_x();
  int y=yy-Fl::event_y();
  angleY-=(double)y/EFFECT; // EFFECT 
  
  if(angleY>90){
    //angleY=90;
    angleY=89.999;
  }else if(angleY<-90){
    //angleY=-90;
    angleY=-89.999;
  }
  angleXZ-=(double)x/EFFECT; // 10 means effect
  CalcCameraPos();
  Reshape();
}

// for perspective window
void
GLWindow::Approach(){
  int x=Fl::event_x();
  int y=Fl::event_y();
  float oldDist=sqrt((xx-viewport[2]/2)*(xx-viewport[2]/2)+
		     (yy-viewport[3]/2)*(yy-viewport[3]/2));
  float nowDist=sqrt((x-viewport[2]/2)*(x-viewport[2]/2)+
		     (y-viewport[3]/2)*(y-viewport[3]/2));  
  cameraR*=oldDist/nowDist;
  CalcCameraPos();
  Reshape();
}

// for perspective and ortho window
void
GLWindow::Pan(){
    Vector3d move,tmp;

    int x=Fl::event_x();
    int y=Fl::event_y();
    
    // for ortho
    GLdouble objX1,objY1,objZ1;
    GLdouble objX2,objY2,objZ2;

    Vector3d vector2,point;
    switch(camera->type){
      //case NORMAL:
    case PERSPECTIVE:
      move=PointOnPlane(vector1,camera->view,xx,yy);
      tmp=PointOnPlane(vector1,camera->view,x,y);
      move.Minus(tmp);    
      camera->camera.Plus(move);
      camera->view.Plus(move);
      CalcCameraPos();
      break;
    default: // for ortho
      gluUnProject(x,camera->height-y,1.0, 
		   modelMatrix,projMatrix,viewport,
		   &objX1,&objY1,&objZ1);
      gluUnProject(xx,camera->height-yy,1.0, 
		   modelMatrix,projMatrix,viewport,
		   &objX2,&objY2,&objZ2);
      move.Set(objX2-objX1,objY2-objY1,objZ2-objZ1);
      camera->camera.Plus(move);
      break;
    }
    Reshape();
}

// for ortho window
void
GLWindow::Zoom(){
  int x=Fl::event_x();
  int y=Fl::event_y();
  float oldDist=sqrt((xx-viewport[2]/2)*(xx-viewport[2]/2)+
		     (yy-viewport[3]/2)*(yy-viewport[3]/2));
  float nowDist=sqrt((x-viewport[2]/2)*(x-viewport[2]/2)+
		     (y-viewport[3]/2)*(y-viewport[3]/2));  
  camera->zoom*=nowDist/oldDist;
  Reshape();
}

void
GLWindow::ReshapeCore(){
  switch(camera->type){
  case PERSPECTIVE:
    gluPerspective(camera->eyeAngle,
		   (GLdouble)camera->width/camera->height,
		   0.01*cameraR,4000);  // TODO
    gluLookAt(camera->camera.X(),camera->camera.Y(),camera->camera.Z(),
	      camera->view.X(),camera->view.Y(),camera->view.Z(),
	      nowUp.X(),nowUp.Y(),nowUp.Z());    
    break;
    //case XY:
  case XYPM:
    if(camera->width < camera->height){
      glOrtho(-1.0/camera->zoom, 
	      1.0/camera->zoom,
	      -1.0*(float)camera->height/camera->width/camera->zoom,
	      1.0*(float)camera->height/camera->width/camera->zoom,
	      minZ-1,maxZ+1);
    } else{
      glOrtho(-1.0*(float)camera->width/camera->height/camera->zoom,
	      1.0*(float)camera->width/camera->height/camera->zoom,
	      -1.0/camera->zoom,
	      1.0/camera->zoom,
	      minZ-1,maxZ+1);     
    }
    glRotatef(180.0, 0, 1.0, 0);
    glTranslatef(-camera->camera.X(),-camera->camera.Y(),0);
    DebugOut("minZ,maxZ==" << minZ << ", " << maxZ);
    break;

  case XYMP:
    if(camera->width < camera->height){
      glOrtho(-1.0/camera->zoom, 
	      1.0/camera->zoom,
	      -1.0*(float)camera->height/camera->width/camera->zoom,
	      1.0*(float)camera->height/camera->width/camera->zoom,
	      -(maxZ+1),-(minZ-1));
    } else{
      glOrtho(-1.0*(float)camera->width/camera->height/camera->zoom,
	      1.0*(float)camera->width/camera->height/camera->zoom,
	      -1.0/camera->zoom,
	      1.0/camera->zoom,
	      -(maxZ+1),-(minZ-1));     
    }
    glTranslatef(-camera->camera.X(),-camera->camera.Y(),0);
    DebugOut("minZ,maxZ==" << minZ << ", " << maxZ);
    break;
    //case YZ:
  case YZMP:
    if(camera->width < camera->height){
      glOrtho(1.0/camera->zoom,
	      -1.0/camera->zoom,
	      -1.0*(float)camera->height/camera->width/camera->zoom,
	      1.0*(float)camera->height/camera->width/camera->zoom,
	      //(maxX+1),(minX-1));
	      -(minX-1),-(maxX+1));
    } else{
      glOrtho(1.0*(float)camera->width/camera->height/camera->zoom, 
	      -1.0*(float)camera->width/camera->height/camera->zoom,
	      -1.0/camera->zoom,
	      1.0/camera->zoom,
	      //(maxX+1),(minX-1));
	      -(minX-1),-(maxX+1));
    }
    glRotatef(-90.0, 0, 1.0, 0);
    glTranslatef(0,-camera->camera.Y(),-camera->camera.Z());
    DebugOut("minX,maxX=" << minX << ", " << maxX);
    break;
  case YZPM:
    if(camera->width < camera->height){
      glOrtho(1.0/camera->zoom,
	      -1.0/camera->zoom,
	      -1.0*(float)camera->height/camera->width/camera->zoom,
	      1.0*(float)camera->height/camera->width/camera->zoom,
	      maxX*2,minX*2);
    } else{
      glOrtho(1.0*(float)camera->width/camera->height/camera->zoom, 
	      -1.0*(float)camera->width/camera->height/camera->zoom,
	      -1.0/camera->zoom,
	      1.0/camera->zoom,
	      maxX*2,minX*2);
    }
    glRotatef(90.0, 0, 1.0, 0);
    glTranslatef(0,-camera->camera.Y(),-camera->camera.Z());
    DebugOut("minX,maxX=" << minX << ", " << maxX);
    break;

    //case XZ:
  case XZMP:
    if(camera->width < camera->height){
      glOrtho(-1.0/camera->zoom, 
	      1.0/camera->zoom,
	      -1.0*(float)camera->height/camera->width/camera->zoom,
	      1.0*(float)camera->height/camera->width/camera->zoom,
	      minY-1,maxY+1);
    } else{
      glOrtho(-1.0*(float)camera->width/camera->height/camera->zoom,
	      1.0*(float)camera->width/camera->height/camera->zoom,
	      -1.0/camera->zoom,
	      1.0/camera->zoom,
	      minY-1,maxY+1);
    }
    glRotatef(-90.0, 1.0, 0, 0);
    glTranslatef(-camera->camera.X(),0,-camera->camera.Z());
    DebugOut("minY,maxY=" << minY << ", " << maxY);
    break;

  case XZPM:
    if(camera->width < camera->height){
      glOrtho(-1.0/camera->zoom, 
	      1.0/camera->zoom,
	      -1.0*(float)camera->height/camera->width/camera->zoom,
	      1.0*(float)camera->height/camera->width/camera->zoom,
	      -(maxY+1),-(minY-1)); //{+,-}1:margin for selected solid vertex
    } else{
      glOrtho(-1.0*(float)camera->width/camera->height/camera->zoom,
	      1.0*(float)camera->width/camera->height/camera->zoom,
	      -1.0/camera->zoom,
	      1.0/camera->zoom,
	      -(maxY+1),-(minY-1));
    }
    glRotatef(90.0, 1.0, 0, 0);
    glTranslatef(-camera->camera.X(),0,-camera->camera.Z());
    DebugOut("minY,maxY=" << minY << ", " << maxY);
    break;
  }
}

void
GLWindow::Reshape() {
  Vector3d z;
  double norm;

  if(!visible_r()){
#ifdef DEBUG
    cout << "can't visible,abort reshape" << endl;    
#endif
    return;
  }
#ifdef DEBUG
  cout << "in Reshape,GLWindow:w(),h()==" << w() << ", " << h() << endl;
#endif
  camera->width=w();
  camera->height=h();

  //cout << "cameraR==" << cameraR << endl;

  switch(camera->type){
  case PERSPECTIVE:
    // Calculate up vector
    vector1=camera->view-camera->camera;
    vector1.Normalize();
    nowUp=vector1;
    //z.Set(0,0,1); // right hand system
    z.Set(0,1,0); // left hand system
    nowUp=vector1.Cross(z);
    nowUp=nowUp.Cross(vector1);
    norm=nowUp.Dot(nowUp);
    norm=sqrt(norm);
    nowUp.Multi(1.0/norm);
    break;
    //case XY:
  case XYPM:
    vector1.coord[0]=0;
    vector1.coord[1]=0;
    vector1.coord[2]=+1;
    break;
  case XYMP:
    vector1.coord[0]=0;
    vector1.coord[1]=0;
    vector1.coord[2]=-1;
    break;
    //case YZ:
  case YZPM:
    vector1.coord[0]=1;
    vector1.coord[1]=0;
    vector1.coord[2]=0;
    break;
  case YZMP:
    vector1.coord[0]=-1;
    vector1.coord[1]=0;
    vector1.coord[2]=0;
    break;
    //case XZ:
  case XZPM:
    vector1.coord[0]=0;
    vector1.coord[1]=1;
    vector1.coord[2]=0;
    break;
  case XZMP:
    vector1.coord[0]=0;
    vector1.coord[1]=-1;
    vector1.coord[2]=0;
    break;
  }
  
  glViewport( 0, 0, camera->width, camera->height );
  glGetIntegerv (GL_VIEWPORT, viewport);
  glMatrixMode( GL_PROJECTION );
  glPushMatrix();
  glLoadIdentity();

  ReshapeCore();

  glGetDoublev(GL_PROJECTION_MATRIX, projMatrix);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glGetDoublev(GL_MODELVIEW_MATRIX, modelMatrix);    
  glPopMatrix();

#ifdef DEBUG
  cout << "projMatrix" << endl;
  for(int cnt1=0;cnt1<4;cnt1++){
    for(int cnt2=0;cnt2<4;cnt2++){
      cout << projMatrix[cnt1*4+cnt2] << " ";
    }
    cout << endl;
  }
  cout << "modelMatrix" << endl;
  for(int cnt1=0;cnt1<4;cnt1++){
    for(int cnt2=0;cnt2<4;cnt2++){
      cout << modelMatrix[cnt1*4+cnt2] << " ";
    }
    cout << endl;
  }
#endif

#ifdef DEBUG
  if(camera->type!=PERSPECTIVE){
    cout << "camera x,y,z == " << camera->camera.X() << ", " 
	 << camera->camera.Y() << ", " 
	 << camera->camera.Z() << endl;
    cout << "zoom=" << camera->zoom << endl;
  }
#endif
}


int
GLWindow::VaguePick(int selectTimes){
  int selectedNum;

#ifdef DEBUG
  cout << "camera num==" << camNum << endl;
  GLfloat* depthBuf;
  depthBuf = new GLfloat[camera->width*camera->height];
  glReadBuffer(GL_FRONT);
  glReadPixels(0,0,camera->width,camera->height,
	       GL_DEPTH_COMPONENT,GL_FLOAT,depthBuf);
  cout  << "get depth info by glReadPixel." << endl;
  FILE* tmpFP;
  tmpFP=fopen("DepthInfo.ppm","w");
  fprintf(tmpFP,"P3\n");
  fprintf(tmpFP,"%d %d\n",camera->width,camera->height);
  fprintf(tmpFP,"255");
  for(int cnt1=0;cnt1<camera->width*camera->height;cnt1++){
    if(cnt1%10 == 0){
      fprintf(tmpFP,"\n");
    }
    fprintf(tmpFP,"%d ",(int)(depthBuf[cnt1]*255.0));
    fprintf(tmpFP,"%d ",(int)(depthBuf[cnt1]*255.0));
    fprintf(tmpFP,"%d ",(int)(depthBuf[cnt1]*255.0));
  }
  fclose(tmpFP);
  delete [] depthBuf;
#endif

  switch(sMode){
  case sVertex:
    if(vertex.size()==0){
      return -1;
    }
    //selectedNum=VertexPick2(selectTimes);
    selectedNum=VertexPick();
    break;
  case sEdge:
    if(edge.size()==0){
      return -1;
    }
    selectedNum=EdgePick();
    break;
  case sFace:
    if(face.size()==0){
      return -1;
    }
    selectedNum=FacePick();
    break;
  }

  return selectedNum;
}

/*
int
GLWindow::SurfacePick3(){
  double mostDist=DBL_MAX;
  Vector3d* tmp;
  GLdouble objX1,objY1,objZ1;
  GLdouble distance,depth;
  
  int selectTimes=11; // 2^11 == 2048 means window size
  GLint hits;
  int pickSize=1;
  
  unsigned int getThisName;
  GLuint names, *ptr;
  GLuint newFace;
  Vector3d pickPoint;
  int x=Fl::event_x();
  int y=Fl::event_y();
  int selectedNum=-1;
  
  // find point near cursor
  for(int cnt1=0;cnt1<subdividedVertex[1];cnt1++){
    tmp=&vertex[cnt1];
    gluProject(tmp->X(),tmp->Y(),tmp->Z(),
	       modelMatrix,projMatrix,viewport,
	       &objX1,&objY1,&objZ1);
    distance=sqrt((xx-objX1)*(xx-objX1)
		  +(camera->height-yy-objY1)*(camera->height-yy-objY1));
    cout << "In Push distance==" << distance << endl;
    cout << "mostDist==" << mostDist << endl;

    if(distance<mostDist){ // found nearer point
      for(int times=0;times<selectTimes;times++){
	pickSize=pickSize*2;
	
	glSelectBuffer (BufSize, selectBuf);
	(void) glRenderMode (GL_SELECT);
	glInitNames();
	glPushName(0);
	
	glViewport( 0, 0, camera->width, camera->height );    
	glMatrixMode (GL_PROJECTION);
	glPushMatrix ();    
	glLoadIdentity ();
	//  create pickSize x pickSize picking region near cursor location    
	gluPickMatrix ((int)objX1, (int)objY1, 
		       pickSize, pickSize, viewport);
	gluPerspective(camera->eyeAngle,
	               (GLdouble)camera->width/camera->height,
		       0.01*cameraR,4000);
	gluLookAt(camera->camera.X(),camera->camera.Y(),camera->camera.Z(),
		  camera->view.X(),camera->view.Y(),camera->view.Z(),
		  nowUp.X(),nowUp.Y(),nowUp.Z());
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();    
		
	for(int cnt2=0;cnt2<subdividedFace[1];cnt2++){
	  glLoadName(cnt2);
	  face[cnt2].Draw();
	}
	glPopMatrix();
	hits = glRenderMode (GL_RENDER);
	cout << "hits==" << hits << endl;
	if(hits!=0){
	  mostDepth=DBL_MAX;
	  ptr=selectBuf;
	  newFace = 0;
	  times=selectTimes;
	  for (int i = 0; i < hits; i++) {  // for each hit 
	    getThisName = 0;
	    names = *ptr;
	    cout << "number of name==" << names << endl;
	    
	    ptr++;              // skip # name
	    cout << "z1==" << *ptr << endl;
	    
	    ptr++;              // skip z1 
	    cout << "z2==" << *ptr << endl;
	    
	    ptr++;              // skip z2 
	    cout << "face name?==" << *ptr << endl;
	    
	    pickPoint=PointOnPlane(face[*ptr].Normal(),
				   vertex[face[*ptr].vertexLink[0]],objX1,objY1);
	    cout << "pick point==" << endl;
	    pickPoint.Print();
	    pickPoint.Minus(camera->camera);
	    depth=pickPoint.Dot(pickPoint);
	    cout << "depth==" << depth << endl;
	    cout << "most Depth==" << mostDepth << endl;
	    
	    if(depth<mostDepth){
	      cout << "change" << endl;
	      mostDepth=depth;
	      getThisName=1;
	    }
	    
	    if (getThisName)
	      newFace = *ptr;
	    ptr += names;       // skip the names list
	  }

	  	  	  
	}

      //cout << "in Push change" << endl;
	mostDist=distance;
	selectedNum=cnt1;
      }
    }
  }
  return selectedNum;
}
*/

int
GLWindow::EdgePick(){
  int selectedNum=-1;
  GLdouble objX1,objY1,objZ1,objX2,objY2,objZ2;
  Vector3d* tmp;
  Vector3d pickPoint;
  double tmpX,tmpY,tmpZ;
  double relative;
  double a,b; // ax + by + c = 0
  //GLuint depthBuf[10];
  GLfloat depthBuf[10];
  GLfloat depth;
  GLdouble distance;
  GLfloat tmpObjZ1;
  //GLdouble tmpDepth;  
  GLfloat tmpDepth;  

  double mostDist=DBL_MAX;
  glReadBuffer(GL_FRONT);
  xx=Fl::event_x();
  yy=Fl::event_y();

  for (int cnt1=0;cnt1<subdividedEdge[1];cnt1++) {  // for each hit     
    tmp=&vertex[edge[cnt1].vertexLink[Left]];
    gluProject(tmp->X(),tmp->Y(),tmp->Z(),
	       modelMatrix,projMatrix,viewport,
	       &objX1,&objY1,&objZ1);
    tmp=&vertex[edge[cnt1].vertexLink[Right]];
    gluProject(tmp->X(),tmp->Y(),tmp->Z(),
	       modelMatrix,projMatrix,viewport,
	       &objX2,&objY2,&objZ2);
    a=objY2-objY1;
    b=objX1-objX2;
    tmpX=-(a*b*(camera->height-yy) - b*b*xx + a*(objX2*objY1-objX1*objY2))
      /(a*a + b*b);
    tmpY=-(a*b*xx - a*a*(camera->height-yy) + b*(objX2*objY1-objX1*objY2))
      /(a*a + b*b);      
    relative=(tmpX-objX1)/(objX2-objX1);
    if(relative>1){
#ifdef DEBUG
      cout << "out of edge" << endl;
#endif
      tmpX=objX2;
      tmpY=objY2;
      relative=1;
   }
    if(relative<0){
#ifdef DEBUG
      cout << "out of edge" << endl;
#endif
      tmpX=objX1;
      tmpY=objY1;
      relative=0;      
    }    
    tmpZ=objZ1+(objZ2-objZ1)*relative;
#ifdef DEBUG
    cout << "tmpX,tmpY,tmpZ==" <<tmpX << ", " <<tmpY << ", " <<tmpZ <<endl;    
    cout << "xx,yy==" << xx << ", " << yy << endl;
#endif
    // out of window
    if(tmpX>=camera->width || tmpX<0 || tmpY>=camera->height || tmpY<0){
#ifdef DEBUG
      cout << "out of window" << endl;
#endif
      continue;
    }

    glReadPixels((int)(tmpX+0.5),(int)(tmpY+0.5),1,1,
		 //GL_DEPTH_COMPONENT,GL_UNSIGNED_INT,depthBuf);
		 GL_DEPTH_COMPONENT,GL_FLOAT,depthBuf);
    // TODO: more accuracy ... muri ka?
    tmpObjZ1=(GLfloat)((GLuint)(tmpZ*0xffffffff)/1000000);
    //tmpDepth=depthBuf[0]/1000000;
    tmpDepth=(GLfloat)((GLuint)(depthBuf[0]*0xffffffff)/1000000);
    tmpObjZ1+=0.5;
    tmpDepth+=0.5;
    tmpObjZ1=(int)tmpObjZ1;
    tmpDepth=(int)tmpDepth;
    
#ifdef DEBUG
    cout << "after:tmpDepth, tmpObjZ1 ==" 
	 << tmpDepth << ", " << tmpObjZ1 << endl;
#endif
    if(tmpDepth+1>=tmpObjZ1 || sFFlag==GL_FALSE){ // front of surface;+1 mergin
      distance=sqrt((xx-tmpX)*(xx-tmpX)+(camera->height-yy-tmpY)*(camera->height-yy-tmpY));
#ifdef DEBUG
      cout << "In EdgePick distance==" << distance << endl;
      cout << "mostDist==" << mostDist << endl;
#endif
      if(distance<mostDist){ // nearest point ever
	//cout << "in SurfacePick change" << endl;
	mostDist=distance;
	selectedNum=cnt1;
      }
    }
  }    
  
  DebugOut("selectedNum==" << selectedNum);
  return selectedNum;
}

int
GLWindow::FacePick(){
  int selectedNum;
  int selectTimes=11; // 2^11 == 2048 will be larger than window size
  GLint hits;
  int pickSize=1;
  int tmpX,tmpY;
  
  tmpX=Fl::event_x();
  tmpY=Fl::event_y();
  
  for(int times=0;times<selectTimes;times++){
    pickSize=pickSize*2;
    
    glSelectBuffer (BufSize, selectBuf);
    (void) glRenderMode (GL_SELECT);
    glInitNames();
    glPushName(0);
    
    glViewport( 0, 0, camera->width, camera->height );    
    glMatrixMode (GL_PROJECTION);
    glPushMatrix ();    
    glLoadIdentity ();
    //  create pickSize x pickSize picking region near cursor location    
    gluPickMatrix ((GLdouble)tmpX, (GLdouble) (viewport[3] - tmpY), 
		   pickSize, pickSize, viewport);

    ReshapeCore();
    /*
  switch(camera->type){
  case NORMAL:
    gluPerspective(camera->eyeAngle,
		   (GLdouble)camera->width/camera->height,
		   0.01*cameraR,4000);
    gluLookAt(camera->camera.X(),camera->camera.Y(),camera->camera.Z(),
	      camera->view.X(),camera->view.Y(),camera->view.Z(),
	      nowUp.X(),nowUp.Y(),nowUp.Z());    
    break;
  case XY:
    if(camera->width < camera->height){
      glOrtho(-1.0/camera->zoom, 
	      1.0/camera->zoom,
	      -1.0*(float)camera->height/camera->width/camera->zoom,
	      1.0*(float)camera->height/camera->width/camera->zoom,
	      minZ*10,maxZ*10);
    } else{
      glOrtho(-1.0*(float)camera->width/camera->height/camera->zoom,
	      1.0*(float)camera->width/camera->height/camera->zoom,
	      -1.0/camera->zoom,
	      1.0/camera->zoom,
	      minZ*10,maxZ*10);      
    }
    glTranslatef(-camera->camera.X(),-camera->camera.Y(),0);
#ifdef DEBUG
    cout << "minZ,maxZ==" << minZ << ", " << maxZ << endl;
#endif
    break;
  case YZ:
    if(camera->width < camera->height){
      glOrtho(1.0/camera->zoom,
	      -1.0/camera->zoom,
	      -1.0*(float)camera->height/camera->width/camera->zoom,
	      1.0*(float)camera->height/camera->width/camera->zoom,
	      minX*10,maxX*10);
    } else{
      glOrtho(1.0*(float)camera->width/camera->height/camera->zoom, 
	      -1.0*(float)camera->width/camera->height/camera->zoom,
	      -1.0/camera->zoom,
	      1.0/camera->zoom,
	      minX*10,maxX*10);
    }
    glRotatef(90.0, 0, 1.0, 0);
    glTranslatef(0,-camera->camera.Y(),-camera->camera.Z());
    break;
  case XZ:
    if(camera->width < camera->height){
      glOrtho(-1.0/camera->zoom, 
	      1.0/camera->zoom,
	      -1.0*(float)camera->height/camera->width/camera->zoom,
	      1.0*(float)camera->height/camera->width/camera->zoom,
	      minZ*10,maxZ*10);
    } else{
      glOrtho(-1.0*(float)camera->width/camera->height/camera->zoom,
	      1.0*(float)camera->width/camera->height/camera->zoom,
	      -1.0/camera->zoom,
	      1.0/camera->zoom,
	      minY*10,maxY*10);      
    }
    glRotatef(90.0, 1.0, 0, 0);
    glTranslatef(-camera->camera.X(),0,-camera->camera.Z());
    break;
  }
  */

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();    
    
    for(int cnt1=0;cnt1<subdividedFace[1];cnt1++){
      glLoadName(cnt1);
      face[cnt1].Draw();
    }
    
    glPopMatrix();
    hits = glRenderMode (GL_RENDER);
    
    if(hits!=0){
      selectedNum=ProcessHits (hits, selectBuf); // {point,edge,face} found
      times=selectTimes;
    }      
  }

  return selectedNum;
}

/*
int
GLWindow::VertexPick2(int selectTimes){
  GLint hits;
  int tmpX,tmpY;  
  int selectedNum=-1;
  int pickSize=1;
  tmpX=Fl::event_x();
  tmpY=Fl::event_y();
  
#ifdef DEBUG
  cout << "selectTimes==" << selectTimes << endl;
#endif
  for(int times=0;times<selectTimes;times++){
    pickSize=pickSize*2;
#ifdef DEBUG
    cout << "pickSize==" << pickSize << endl;
#endif

    glRenderMode (GL_RENDER); // temp
    //glSelectBuffer (BufSize, selectBuf);
    //(void) glRenderMode (GL_SELECT);
    
    glViewport( 0, 0, camera->width, camera->height );    
    glMatrixMode (GL_PROJECTION);
    glPushMatrix ();    
    glLoadIdentity ();
    //  create pickSize x pickSize picking region near cursor location    
    gluPickMatrix ((GLdouble)tmpX, 
		   (GLdouble) (viewport[3] - tmpY), 
		   pickSize, pickSize, viewport);
    gluPerspective(camera->eyeAngle,
		   (GLdouble)camera->width/camera->height,
		   0.01*cameraR,4000);
    gluLookAt(camera->camera.X(),camera->camera.Y(),camera->camera.Z(),
	      camera->view.X(),camera->view.Y(),camera->view.Z(),
	      nowUp.X(),nowUp.Y(),nowUp.Z());
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();    

#ifdef DEBUG
    // test matrix
    GLdouble tmpMatrix[16];
    glGetDoublev(GL_PROJECTION_MATRIX, tmpMatrix);
    cout << "projMat" << endl;
    for(int cnt1=0;cnt1<16;cnt1++){
      cout << tmpMatrix[cnt1] << " ";
    }
    cout << endl;
    glGetDoublev(GL_MODELVIEW_MATRIX, tmpMatrix);
    cout << "modelMat" << endl;
    for(int cnt1=0;cnt1<16;cnt1++){
      cout << tmpMatrix[cnt1] << " ";
    }
    cout << endl;
#endif
    if(drawFace==GL_TRUE){
#ifdef DEBUG
      cout << "drawing faces in VertexPick2.." << endl;
#endif
      for(int cnt1=subdividedFace[subdivideLevel];
	  cnt1<subdividedFace[subdivideLevel+1];cnt1++){
	face[cnt1].Draw();
      }
    }else{
#ifdef DEBUG
      cout << "skip drawing faces in VertexPick2.." << endl;
#endif
    }
#ifdef DEBUG
    for(int cnt1=0;cnt1<15;cnt1++){
      selectBuf[cnt1]=-1;
    }
#endif
    glSelectBuffer (BufSize, selectBuf);
    (void) glRenderMode (GL_SELECT);
#ifdef DEBUG
    // test matrix
    glGetDoublev(GL_PROJECTION_MATRIX, tmpMatrix);
    cout << "projMat" << endl;
    for(int cnt1=0;cnt1<16;cnt1++){
      cout << tmpMatrix[cnt1] << " ";
    }
    cout << endl;
    glGetDoublev(GL_MODELVIEW_MATRIX, tmpMatrix);
    cout << "modelMat" << endl;
    for(int cnt1=0;cnt1<16;cnt1++){
      cout << tmpMatrix[cnt1] << " ";
    }
    cout << endl;
#endif

    glEnable(GL_DEPTH_TEST);

    glInitNames();
    glPushName(0);
    GLUquadricObj *sphere;
    sphere = gluNewQuadric();
    double sphereR=10*cameraR*0.001;

    for(int cnt1=0;cnt1<subdividedVertex[1];cnt1++){
      glLoadName(cnt1);
      glPushMatrix();
      glTranslatef(vertex[cnt1].X(),vertex[cnt1].Y(),vertex[cnt1].Z());
      gluSphere(sphere, sphereR, 5, 5); // TODO: radius,tess
      glPopMatrix();
    }
    
    glPopMatrix();
    hits = glRenderMode (GL_RENDER);
     
    if(hits!=0){
      selectedNum=ProcessHits2 (hits, selectBuf); // {point,edge,face} found
      if(selectedNum!=-1){
	break;
      }
    }      
  }
  return selectedNum;
  //return -1;
}
*/

int
GLWindow::VertexPick(){
  //int selectTimes=11; // 2^11 == 2048 means window size
  //GLint hits;
  //int pickSize=1;
  //int tmpX,tmpY;
  //int selectedNum;
  
  /*
  tmpX=Fl::event_x();
  tmpY=Fl::event_y();
  
  for(int times=0;times<selectTimes;times++){
    pickSize=pickSize*2;
    
    glSelectBuffer (BufSize, selectBuf);
    (void) glRenderMode (GL_SELECT);
    glInitNames();
    glPushName(0);
    
    glViewport( 0, 0, camera->width, camera->height );    
    glMatrixMode (GL_PROJECTION);
    glPushMatrix ();    
    glLoadIdentity ();
    //  create pickSize x pickSize picking region near cursor location    
    gluPickMatrix ((GLdouble)tmpX, (GLdouble) (viewport[3] - tmpY), 
		   pickSize, pickSize, viewport);
    gluPerspective(camera->eyeAngle,
		   (GLdouble)camera->width/camera->height,
		   0.01*cameraR,4000);
    gluLookAt(camera->camera.X(),camera->camera.Y(),camera->camera.Z(),
	      camera->view.X(),camera->view.Y(),camera->view.Z(),
	      nowUp.X(),nowUp.Y(),nowUp.Z());
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();    
    
    for(int cnt1=0;cnt1<subdividedFace[1];cnt1++){
      glLoadName(cnt1);
      face[cnt1].Draw();
    }
    
    glPopMatrix();
    hits = glRenderMode (GL_RENDER);
     
    if(hits!=0){
      selectedNum=ProcessHits (hits, selectBuf); // {point,edge,face} found
      break;
    }      
  }

  */


  //GLuint depthBuf[10];
  GLfloat depthBuf[10];
  Vector3d* tmp;
  GLdouble objX1,objY1,objZ1;

  GLfloat tmpObjZ1;
  //GLdouble tmpDepth;
  GLfloat tmpDepth;
  double mostDist=DBL_MAX;
  double distance;
  int selectedNum=-1;
  glReadBuffer(GL_FRONT);

  Vector3d extra;
  double extraX=0;
  double extraY=0;
  double extraZ=0;
  switch(camera->type){
    //case NORMAL:
  case PERSPECTIVE:
    extra=camera->camera-camera->view;
    extraX=extra.X()*0.01; // temp
    extraY=extra.Y()*0.01; // temp
    extraZ=extra.Z()*0.01; // temp
    break;
    //case XY:
  case XYPM:
    extraZ=-0.01/camera->zoom*ExtraR;
    break;
  case XYMP:
    extraZ=0.01/camera->zoom*ExtraR;
    break;
    //case YZ:
  case YZPM:
    extraX=0.01/camera->zoom*ExtraR;
    break;
  case YZMP:
    extraX=-0.01/camera->zoom*ExtraR;
    break;
    //case XZ:
  case XZPM:
    extraY=0.01/camera->zoom*ExtraR;
    break;
  case XZMP:
    extraY=-0.01/camera->zoom*ExtraR;
    break;
  }
  
  for(int cnt1=0;cnt1<subdividedVertex[1];cnt1++){
#ifdef DEBUG
    cout << "vertex[" << cnt1 << "]" << endl;
#endif
    tmp=&vertex[cnt1];
    gluProject(tmp->X()+extraX,tmp->Y()+extraY,tmp->Z()+extraZ,
	       modelMatrix,projMatrix,viewport,
	       &objX1,&objY1,&objZ1);
    // out of window
    if(objX1>=camera->width || objX1<0 || objY1>=camera->height || objY1<0){
#ifdef DEBUG
      cout << "out of window" << endl;
#endif
      continue;
    }
    if((xx-objX1)*(xx-objX1)+       
       // 100 means threshould
       (camera->height-yy-objY1)*(camera->height-yy-objY1)>100 &&
       !Fl::event_state(FL_CTRL)){ // Connect Mode
      continue;
    }
    glReadPixels((int)(objX1+0.5),(int)(objY1+0.5),1,1,
		 //GL_DEPTH_COMPONENT,GL_UNSIGNED_INT,depthBuf);
		 GL_DEPTH_COMPONENT,GL_FLOAT,depthBuf);

#ifdef DEBUG
    cout << "before:depthBuf, objZ1 ==" 
	 << depthBuf[0] << ", " << objZ1 << endl;
#endif

    /*
    // TODO: more accuracy ... muri ka?
    tmpObjZ1=(GLfloat)((GLuint)(objZ1*0xffffffff)/1000000);  
    tmpDepth=(GLfloat)((GLuint)(depthBuf[0]*0xffffffff)/1000000);
    */

    /*
    tmpObjZ1+=0.5;
    tmpDepth+=0.5;
    tmpObjZ1=(int)tmpObjZ1;
    tmpDepth=(int)tmpDepth;
    */
    /*
#ifdef DEBUG
    cout << "after:tmpDepth, tmpObjZ1 ==" 
	 << tmpDepth << ", " << tmpObjZ1 << endl;
#endif
*/
    objY1=camera->height-objY1;
    if(depthBuf[0]>=objZ1 || sFFlag==GL_FALSE){ // front of surface
      distance=sqrt((xx-objX1)*(xx-objX1)+(yy-objY1)*(yy-objY1));
#ifdef DEBUG
      cout << "In SurfacePick distance==" << distance << endl;
      cout << "mostDist==" << mostDist << endl;
#endif
      if(distance<mostDist){ // nearest point ever
	//cout << "in SurfacePick change" << endl;
	mostDist=distance;
	selectedNum=cnt1;
      }
    }
  } 

  DebugOut("selectedNum==" << selectedNum);
  return selectedNum;
}


/*
int
GLWindow::SurfacePick2(){
  int selectTimes=11; // 2^11 == 2048 means window size
  GLint hits;
  int pickSize=1;
  int tmpX,tmpY;
  
  tmpX=Fl::event_x();
  tmpY=Fl::event_y();

  for(int times=0;times<selectTimes;times++){
    pickSize=pickSize*2;
    
    glSelectBuffer (BufSize, selectBuf);
    (void) glRenderMode (GL_SELECT);
    glInitNames();
    glPushName(0);
    
    glViewport( 0, 0, camera->width, camera->height );    
    glMatrixMode (GL_PROJECTION);
    glPushMatrix ();    
    glLoadIdentity ();
    //  create pickSize x pickSize picking region near cursor location    
    gluPickMatrix ((GLdouble)tmpX, (GLdouble) (viewport[3] - tmpY), 
		   pickSize, pickSize, viewport);
    gluPerspective(camera->eyeAngle,
		   (GLdouble)camera->width/camera->height,
		   0.01*cameraR,4000);
    gluLookAt(camera->camera.X(),camera->camera.Y(),camera->camera.Z(),
	      camera->view.X(),camera->view.Y(),camera->view.Z(),
	      nowUp.X(),nowUp.Y(),nowUp.Z());
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();    
    

      for(int cnt1=0;cnt1<subdividedFace[1];cnt1++){
	glLoadName(cnt1);
	face[cnt1].Draw();
      }

      glPopMatrix();
      hits = glRenderMode (GL_RENDER);

      if(hits!=0){
	return ProcessHits (hits, selectBuf); // {point,edge,face} found
      }      
    }
  return -1; // no {point,edge,face} found
}
*/


int
GLWindow::ProcessHits2(GLint hits, GLuint buffer[]){
  unsigned int i, getThisName;
  GLuint names, *ptr;
  //GLuint newSelect;
  GLint newSelect;
  Vector3d pickPoint;
  double dist;
  double mostDist=DBL_MAX;
  int x=Fl::event_x(); // mouse x
  int y=Fl::event_y(); // mouse y
  GLdouble objX1,objY1,objZ1; // for projection
  
  ptr = (GLuint *) buffer;
  newSelect = -1;

#ifdef DEBUG
  cout << "selection buffer.." << endl;
  for(int cnt1=0;cnt1<15;cnt1++){
    cout << buffer[cnt1] << endl;
  }
#endif
  
#ifdef DEBUG
  cout << "hits==" << hits << endl;
#endif
  for (i = 0; i < hits; i++) {  // for each hit 
    getThisName = 0;
    names = *ptr;
    if(names==0){
      //return -1; // picked drawnface
      ptr++;
      ptr++;
      ptr++;
      continue;
    }
    cout << "number of name==" << names << endl;
    
    ptr++;              // skip # name
    cout << "z1==" << *ptr << endl;
    ptr++;              // skip z1 
    cout << "z2==" << *ptr << endl;    
    ptr++;              // skip z2 
    cout << "name?==" << *ptr << endl;
    gluProject(vertex[names].X(),vertex[names].Y(),vertex[names].Z(),
	       modelMatrix,projMatrix,viewport,
	       &objX1,&objY1,&objZ1);
#ifdef DEBUG
    cout << "mouse x,y == " << x << ", " << y << endl;
    cout << "obj x,y == " << objX1 << ", " << objY1 << endl;
#endif
    objY1=camera->height-objY1;
#ifdef DEBUG
    cout << "inverte objY1" << endl;
#endif
    dist=sqrt((xx-objX1)*(xx-objX1)+(yy-objY1)*(yy-objY1));
#ifdef DEBUG
    cout << "In Push distance==" << dist << endl;
#endif    

    if(dist<mostDist){
#ifdef DEBUG
      cout << "change" << endl;
#endif
      mostDepth=dist;
      getThisName=1;
    }
    
    if (getThisName){
      newSelect = *ptr;
    }
    ptr += names;       // skip the names list
  }
  
#ifdef DEBUG
  cout << "newSelect==" << newSelect << endl;
#endif
  return newSelect;
}


// return selected face number
// select exactly hit face
int
GLWindow::ProcessHits (GLint hits, GLuint buffer[]){
  unsigned int i, getThisName;
  GLuint names, *ptr;
  GLuint newSelect;
  Vector3d pickPoint;
  double distance;
  

  ptr = (GLuint *) buffer;
  newSelect = 0;
  mostDepth=DBL_MAX;
  
  if(sMode==sFace){
    for (i = 0; i < hits; i++) {  // for each hit 
      getThisName = 0;
      names = *ptr;
      //cout << "number of name==" << names << endl;
      
      ptr++;              // skip # name
      //cout << "z1==" << *ptr << endl;
      
      ptr++;              // skip z1 
      //cout << "z2==" << *ptr << endl;
      
      ptr++;              // skip z2 
      //cout << "face name?==" << *ptr << endl;
      
      pickPoint=PointOnPlane(face[*ptr].Normal(),
			     vertex[face[*ptr].vertexLink[0]],xx,yy);
      //cout << "pick point==" << endl;
      //pickPoint.Print();
      pickPoint.Minus(camera->camera);
      distance=pickPoint.Dot(pickPoint);
#ifdef DEBUG
      cout << "distance==" << distance << endl;
      cout << "most Depth==" << mostDepth << endl;
#endif
      
      if(distance<mostDepth){
#ifdef DEBUG
	cout << "change" << endl;
#endif
	mostDepth=distance;
	getThisName=1;
      }
      
      if (getThisName)
	newSelect = *ptr;
      ptr += names;       // skip the names list
    }
  }

  if(sMode==sEdge){
  }
  
#ifdef DEBUG
  cout << "newSelect==" << newSelect << endl;
#endif
  return newSelect;
}
/*
  // newFace is selected face number

  Vector3d* tmp;
  double objX1,objY1,objZ1;
  double objX2,objY2,objZ2;
  //int selectedNum;
  Edge* tmpE;
  int selectedNum;
  mostDepth=DBL_MAX;

  switch(sMode){
  case sFace:
    selectedNum=newFace;
    break;
  case sEdge:
    for(int cnt1=0;cnt1<face[newFace].edgeLink.size();cnt1++){
      tmpE=&edge[face[newFace].edgeLink[cnt1]];
      tmp=&vertex[tmpE->vertexLink[Left]];
      gluProject(tmp->X(),tmp->Y(),tmp->Z(),
		 modelMatrix,projMatrix,viewport,
		 &objX1,&objY1,&objZ1);      
      objY1=camera->height-objY1;
      tmp=&vertex[tmpE->vertexLink[Right]];
      gluProject(tmp->X(),tmp->Y(),tmp->Z(),
		 modelMatrix,projMatrix,viewport,
		 &objX2,&objY2,&objZ2);   
      objY2=camera->height-objY2;
      // distance between a point and a line
      distance=abs((objY2-objY1)*xx-(objX2-objX1)*yy+objX2*objY1-objX1*objY2)
	/sqrt((objY2-objY1)*(objY2-objY1)+(objX2-objX1)*(objX2-objX1));
      cout << "distance in edgepick==" << distance << endl;
      if(distance<mostDepth){
	cout << "in Push change" << endl;
	mostDepth=distance;
	selectedNum=face[newFace].edgeLink[cnt1];      
      }
    }
    break;

  case sVertex:
    for(int cnt1=0;cnt1<face[newFace].vertexLink.size();cnt1++){
      cout << "vertex[" << face[newFace].vertexLink[cnt1]
	   << "]" << endl;
      tmp=&vertex[face[newFace].vertexLink[cnt1]];
      gluProject(tmp->X(),tmp->Y(),tmp->Z(),
		 modelMatrix,projMatrix,viewport,
		 &objX1,&objY1,&objZ1);
      cout << "xx,yy,objX1,objY1==" << xx << "," << yy << "," 
	   << objX1 << "," << objY1 << endl;
      objY1=camera->height-objY1;
      cout << "inverted objY1" << endl;
      distance=sqrt((xx-objX1)*(xx-objX1)+(yy-objY1)*(yy-objY1));
      cout << "In Push distance==" << distance << endl;
      cout << "mostDepth==" << mostDepth << endl;
      if(distance<mostDepth){
	cout << "in Push change" << endl;
	mostDepth=distance;
	selectedNum=face[newFace].vertexLink[cnt1];
      }
    }    
    break;
  }  
  return selectedNum;
}
*/

/*
int
GLWindow::ProcessHits (GLint hits, GLuint buffer[]){
  GLuint names, *ptr;

  ptr = (GLuint *) buffer;
  names = *ptr;
  ptr++;

  ptr++;

  ptr++;
  return *ptr;
}
*/

void
GLWindow::SelectMode(int x,int y,int selectTimes){
    vector<int>::iterator where;
    vector<int>* select;

    switch(sMode){
    case sVertex:
      select=&selectV;
      break;
    case sEdge:
      select=&selectE;
      break;
    case sFace:
      select=&selectF;
      break;
    }
    nowV=VaguePick(selectTimes); // get object index
    where=std::find(select->begin(),select->end(),nowV); // find int stl
    if(where==select->end()){ // not found
      select->push_back(nowV);
    }else { // found
      select->erase(where);
    }
    RefineTotalV();
}

// flag specify generate vertex tmporaly(true) or not(false)
void
GLWindow::ConnectMode(int x,int y,int selectTimes,GLboolean flag){
    nowV=VaguePick(selectTimes);
    DebugOut("nowV:" << nowV);
    
    if(flag==GL_FALSE){ // temporary
      if(nowV>=0){ // pick existing vertex
	newV=vertex[nowV];
	fl_cursor(FL_CURSOR_CROSS,FL_WHITE,FL_BLACK);	
      }else{
	newV=Vertex(PointOnPlane(vector1,PassP(),x,y));
	fl_cursor(FL_CURSOR_DEFAULT,FL_WHITE,FL_BLACK);	      
      }
      return;
    }
    if(selectV.size()==0){ // not yet selected point
	if(nowV>=0){ // pick up
	  selectV.push_back(nowV);	    
	}else{ // no point
	  nowV=subdividedVertex[1];
	  vertex.push_back(Vertex(PointOnPlane(vector1,PassP(),x,y)));//no subdivide
	  subdividedVertex[1]++;
	  // check max,min
	  CheckBB(vertex.back());

	  SetChanged(1);
	  selectV.push_back(subdividedVertex[1]-1);
	}
    }else { // already selected
	if(nowV<0){ // no point
	  vertex.push_back(Vertex(PointOnPlane(vector1,PassP(),x,y)));
	  subdividedVertex[1]++;
	  // check max,min
	  CheckBB(vertex.back());
	  SetChanged(1);
	  nowV=subdividedVertex[1]-1;
	}
	for(int cnt1=0;cnt1<selectV.size();cnt1++){
	    AddEdge(selectV[cnt1],nowV);	      
	}
	selectV.erase(selectV.begin(),selectV.end());
	selectV.push_back(nowV); 
    }
    RefineTotalV();
}


const Vector3d
GLWindow::PointOnPlane(Vector3d planeN,Vector3d passPoint,int x,int y){ // plane normal
  GLdouble objX1,objY1,objZ1;
  Vector3d vector2,point,passP;
  passP=passPoint;
  gluUnProject(x,camera->height-y,0.99, 
	       modelMatrix,projMatrix,viewport,
	       &objX1,&objY1,&objZ1);
  switch(camera->type){
  case PERSPECTIVE:
    vector2.Set(objX1,objY1,objZ1);
    vector2.Minus(camera->camera);
    
    double Vx,Vy,Vz,Cx,Cy,Cz,Nx,Ny,Nz,d;
    Vx=vector2.coord[0];
    Vy=vector2.coord[1];
    Vz=vector2.coord[2];
    Cx=camera->camera.coord[0];
    Cy=camera->camera.coord[1];
    Cz=camera->camera.coord[2];
    Nx=planeN.coord[0];
    Ny=planeN.coord[1];
    Nz=planeN.coord[2];
    d=-Nx*passP.coord[0]-Ny*passP.coord[1]-Nz*passP.coord[2];
    if(vector2.coord[0]==0){
      point.coord[0]=camera->camera.coord[0];
    }else{
      point.coord[0]=(Vy/Vx*Cx*Ny+Vz/Vx*Cx*Nz-Cy*Ny-Cz*Nz-d)/
	(Nx+Vy/Vx*Ny+Vz/Vx*Nz);
    }
    if(vector2.coord[1]==0){
      point.coord[1]=camera->camera.coord[1];
    }else{
      point.coord[1]=(Vx/Vy*Cy*Nx+Vz/Vy*Cy*Nz-Cx*Nx-Cz*Nz-d)/
	(Ny+Vx/Vy*Nx+Vz/Vy*Nz);
    }
    if(vector2.coord[2]==0){
      point.coord[2]=camera->camera.coord[2];
    }else{
      point.coord[2]=(Vx/Vz*Cz*Nx+Vy/Vz*Cz*Ny-Cy*Ny-Cx*Nx-d)/
	(Nz+Vx/Vz*Nx+Vy/Vz*Ny);
    }
    break;
  case XYPM:
  case XYMP:
    point.Set(objX1,objY1,passP.Z());
    break;
  case YZPM:
  case YZMP:
    point.Set(passP.X(),objY1,objZ1);
    break;
  case XZPM:
  case XZMP:
    point.Set(objX1,passP.Y(),objZ1);
    break;
  }
  return point;
}


void
Identity(double* mat){
  for(int cnt1=0;cnt1<9;cnt1++){
    mat[cnt1]=0;
  }
  mat[0]=1;
  mat[4]=1;
  mat[8]=1;
}

void
CalcMat33(double* mat,Vector3d& vec,double angle){
  double cosR=cos(0.05*angle/EFFECT);
  double sinR=sin(0.05*angle/EFFECT);
  mat[0]=vec.coord[0]*vec.coord[0]*(1-cosR)+cosR;
  mat[1]=vec.coord[0]*vec.coord[1]*(1-cosR)-vec.coord[2]*sinR;
  mat[2]=vec.coord[0]*vec.coord[2]*(1-cosR)+vec.coord[1]*sinR;
  mat[3]=vec.coord[1]*vec.coord[0]*(1-cosR)+vec.coord[2]*sinR;
  mat[4]=vec.coord[1]*vec.coord[1]*(1-cosR)+cosR;
  mat[5]=vec.coord[1]*vec.coord[2]*(1-cosR)-vec.coord[0]*sinR;
  mat[6]=vec.coord[2]*vec.coord[0]*(1-cosR)-vec.coord[1]*sinR;
  mat[7]=vec.coord[2]*vec.coord[1]*(1-cosR)+vec.coord[0]*sinR;
  mat[8]=vec.coord[2]*vec.coord[2]*(1-cosR)+cosR;
}


// TODO 
// reflect user settings
void
GLWindow::Init( void ){
    /* setup lighting, etc */
    mode(FL_RGB|FL_DOUBLE|FL_DEPTH);

    newFlag=GL_FALSE;

    glClearColor(0.4, 0.4, 0.4, 1.0);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );    
//#ifdef DEBUG
    if(glGetError()==GL_INVALID_VALUE){
      cout << "GL_INVALID_VALUE" << endl;
    }
    if(glGetError()==GL_INVALID_ENUM){
      cout << "GL_INVALID_ENUM" << endl;
    }
    if(glGetError()==GL_INVALID_OPERATION){
      cout << "GL_INVALID_OPERATION" << endl;
    }
    if(glGetError()==GL_STACK_OVERFLOW){
      cout << "GL_STACK_OVERFLOW" << endl;
    }
    if(glGetError()==GL_STACK_UNDERFLOW){
      cout << "GL_STACK_UNDERFLOW" << endl;
    }
    if(glGetError()==GL_OUT_OF_MEMORY){
      cout << "GL_OUT_OF_MEMORY" << endl;
    }
//#endif

    glColor3f(1.0, 1.0, 1.0);
    glLightfv(GL_LIGHT0,GL_POSITION,light0_pos);
    glLightfv(GL_LIGHT0,GL_DIFFUSE,light0_dif);
    glLightfv(GL_LIGHT1,GL_POSITION,light1_pos);
    glLightfv(GL_LIGHT1,GL_DIFFUSE,light1_dif);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,1);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material_dif0); 
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material_amb0); 
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHT1);
    glEnable(GL_NORMALIZE);
    glShadeModel(GL_FLAT);
    glDepthFunc(GL_LEQUAL);
    //glDepthFunc(GL_LESS);
    glPointSize(PointSize);
    glLineWidth(1);
    glEnable(GL_DEPTH_TEST);
    glDepthRange(0.0, 1.0); 
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);    
}

GLboolean
Toggle(GLboolean& flag){
  if (flag==GL_TRUE){
    flag=GL_FALSE;
  }else {
    flag=GL_TRUE;
  }
  return flag;
  //  flag=!flag; // why not work?
}

void
ToggleX(){
  if(Toggle(xConst)==GL_FALSE){
    mainM->restrictX->clear();
    restX->clear();
  }else{
    mainM->restrictX->set();
    restX->set();
  }
  Fl::redraw();
}


void
ToggleY(){
  if(Toggle(yConst)==GL_FALSE){
    mainM->restrictY->clear();
    restY->clear();
  }else{
    mainM->restrictY->set();
    restY->set();
  }
  Fl::redraw();
}

void
ToggleZ(){
  if(Toggle(zConst)==GL_FALSE){
    mainM->restrictZ->clear();
    restZ->clear();
  }else{
    mainM->restrictZ->set();
    restZ->set();
  }
  Fl::redraw();
}

int
GLWindow::Key(){
  int nowKey;
  nowKey=Fl::event_key();
  // return value {0,1,2}: 0 no mean,1 redraw this window,2 redraw all window
  int value=0; 

  switch (nowKey) {
  case 'w':
    if(Fl::event_state(FL_SHIFT)){
      SelectedSharp();
      WeightPropagate();
      for(int cnt1=0;cnt1<subdivideLevel;cnt1++){
	RecalcSub(cnt1);
      }
      value=2;
      break;
    }
    SelectedSmooth();
    WeightPropagate();
    for(int cnt1=0;cnt1<subdivideLevel;cnt1++){
      RecalcSub(cnt1);
    }
    value=2;
    break;    
  default:
    return 0;
  }
  if(value==1){
    redraw();    
  }else if(value==2){
    Fl::redraw();
  }
  return value;
}

void
GLWindow::DrawAxe(){
  glDisable(GL_COLOR);
  float scale;
  if(camera->type==PERSPECTIVE){
    scale=cameraR;
  }else{
    scale=1.0/camera->zoom;
  }
  glScalef(scale,scale,scale);
  // X Axe
  glColor3f((float)(1.0-0.5*xConst),0,0);
  glBegin(GL_LINES);
  glVertex3f(0,0,0);
  glVertex3f(10,0,0);
  glEnd();
  // Y Axe
  glColor3f(0,(float)(1.0-0.5*yConst),0);
  glBegin(GL_LINES);
  glVertex3f(0,0,0);
  glVertex3f(0,10,0);
  glEnd();
  // Z Axe
  glColor3f(0,0,(float)(1.0-0.5*zConst));
  glBegin(GL_LINES);
  glVertex3f(0,0,0);
  glVertex3f(0,0,-10); // trick for left hand system
  glScalef(1.0/scale,1.0/scale,1.0/scale);
  glEnd();  
}

void
GLWindow::Display( void ){
  DebugOut("Debug Out in " << camNum << "th Display");

  glRenderMode(GL_RENDER);
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  
  //glLoadIdentity();
  //glMatrixMode(GL_MODELVIEW);
  //glLoadIdentity();
  glEnable(GL_LIGHTING);  

  // Face
  if(sFFlag && fFlag){
    int subStart=subdividedFace[subdivideLevel];
    int subEnd=subdividedFace[subdivideLevel+1];
    for(int cnt1=subStart;cnt1<subEnd;cnt1++){
      face[cnt1].Draw();
    }
  }
  glDisable(GL_LIGHTING);

  // Edge
  glLineWidth(1);  
  if(eFlag==GL_TRUE){
    glColor3f(0.6, 0.8, 0.7);
    int end0=subdividedEdge[1]; // end order in subdivide level 0
    for(int cnt1=0;cnt1<end0;cnt1++){
      edge[cnt1].Draw();
    }
  }
  if(sEFlag && fFlag){
    glColor3f(0.9, 0.9, 0.9);
    int subStart=subdividedEdge[subdivideLevel];
    int subEnd=subdividedEdge[subdivideLevel+1];
    for(int cnt1=subStart;cnt1<subEnd;cnt1++){
      edge[cnt1].Draw();
    }
  }

  // Vertex
  glPointSize(PointSize);
  glColor3f(0.9, 0.7, 0.9);
  // original vertices
  if(vFlag==GL_TRUE){
    glBegin(GL_POINTS); // temp
    int end0=subdividedVertex[1]; // end order in subdivide level 0
    for(int cnt1=0;cnt1<end0;cnt1++){
      vertex[cnt1].Draw(); // temp
    }
    glEnd();
  }
  // subdivided vertices    
  if(sVFlag && fFlag){
    glPointSize(1.0);
    glColor3f(0.9, 0.9, 0.9);
    int subStart=subdividedVertex[subdivideLevel];
    int subEnd=subdividedVertex[subdivideLevel+1];
    glBegin(GL_POINTS); 
    for(int cnt1=subStart;cnt1<subEnd;cnt1++){
      vertex[cnt1].Draw();
    }
    glEnd();
  }

  // selected object ///////////////////////

  // selected Face
  glColor3f(0.7, 0.4, 0.1);
  for(int cnt1=0;cnt1<selectF.size();cnt1++){
     face[selectF[cnt1]].Draw();
  }

  // selected Edge
  //glLineWidth(3.0);
  glColor3f(0.1, 0.1, 0.8);
  for(int cnt1=0;cnt1<selectE.size();cnt1++){
    edge[selectE[cnt1]].Draw();
  }

  // selected Vertex
  //glPointSize(8.0);
  glColor3f(0.0, 1.0, 0.0); // BLUE
  //glBegin(GL_POINTS);
  int selectVSize =selectV.size();

  GLUquadricObj *sphere;
  sphere = gluNewQuadric();
  double sphereR;
  if(camera->type==PERSPECTIVE){
    sphereR=10*cameraR*0.001;
  }else{
    sphereR=0.01/camera->zoom*ExtraR;
  }

  for(int cnt1=0;cnt1<selectVSize-1;cnt1++){
    //vertex[selectV[cnt1]].Draw();
    glPushMatrix();
    glTranslatef(vertex[selectV[cnt1]].X(),
		 vertex[selectV[cnt1]].Y(),
		 vertex[selectV[cnt1]].Z());
    gluSphere(sphere, sphereR, 4, 4); // TODO: radius,tess
    glPopMatrix();
  }
  glColor3f(1.0, 0.0, 0.0); // RED
  if(selectV.size()>0){
    //vertex[selectV[selectVSize-1]].Draw();
    glPushMatrix();
    glTranslatef(vertex[selectV[selectVSize-1]].X(),
		 vertex[selectV[selectVSize-1]].Y(),
		 vertex[selectV[selectVSize-1]].Z());
    gluSphere(sphere, sphereR, 4, 4); // TODO: radius,tess
    glPopMatrix();
  }
  //glEnd();

  // temporary objects
  if(newFlag==GL_TRUE){
    glColor3f(0.7, 0.0, 0.0); // RED
    glPushMatrix();
    glTranslatef(newV.X(),
		 newV.Y(),
		 newV.Z());
    gluSphere(sphere, sphereR, 4, 4); // TODO: radius,tess
    glPopMatrix();    
  }
  gluDeleteQuadric(sphere);
  

  // {X,Y,Z} axe
  glLineWidth(1);  
  if(axeFlag==GL_TRUE){
    DrawAxe();
  }
}

void
GLWindow::ToggleAxe(GLboolean value){
  axeFlag=value;
  redraw();
}

void
GLWindow::ToggleVertex(GLboolean value){
  vFlag=value;
  //sVFlag=value;
  redraw();
}

void
GLWindow::ToggleEdge(GLboolean value){
  eFlag=value;
  //sEFlag=value;
  redraw();
}

void
GLWindow::ToggleFace(GLboolean value){
  fFlag=value;
  //sFFlag=value;
  redraw();
}

void
GLWindow::ViewTarget(){
  CheckBBS();
  Vector3d center;
  center.Set((minXS+maxXS)/2,(minYS+maxYS)/2,(minZS+maxZS)/2);    
  if(camera->type==PERSPECTIVE){
    camera->view=center;
    CalcCameraPos();
  }else{
    camera->camera=center;
  }
  redraw();
}

void
GLWindow::ViewWhole(){
  for(int cnt1=0;cnt1<subdividedVertex[1];cnt1++){
    CheckBBSCore(vertex[cnt1]);
  }
  Vector3d center;
  center.Set((minXS+maxXS)/2,(minYS+maxYS)/2,(minZS+maxZS)/2);    
  double diagonal=sqrt((maxXS-minXS)*(maxXS-minXS))
    +sqrt((maxYS-minYS)*(maxYS-minYS))
    +sqrt((maxZS-minZS)*(maxZS-minZS));
  switch(camera->type){
  case PERSPECTIVE:
    camera->view=center;
    CalcCameraPos();
    break;
  case XYPM:
  case XYMP:
    if(camera->width < camera->height){
      
    }else{

    }
    break;
  case YZPM:
  case YZMP:
    break;
  case XZPM:
  case XZMP:
    break;
  }
  if(camera->type!=PERSPECTIVE){
    camera->camera=center;
  }
}

void
GLWindow::SVDraw(){
  sVFlag=GL_TRUE;
  sEFlag=GL_FALSE;
  sFFlag=GL_FALSE;
}

void
GLWindow::SEDraw(){
  sVFlag=GL_FALSE;
  sEFlag=GL_TRUE;
  sFFlag=GL_FALSE;
}

void
GLWindow::SFDraw(){
  sVFlag=GL_FALSE;
  sEFlag=GL_FALSE;
  sFFlag=GL_TRUE;
}


