#ifndef _PYGMALION_H_
#define _PYGMALION_H_
#include <FL/Fl.H>
#include <FL/Fl_Gl_Window.H>
#include <FL/Fl_Menu_Item.H>
#include <FL/gl.h>

#include "camera.h"
 
#ifdef DEBUG
#define DebugOut(a); cout << a << endl;
#else
#define DebugOut(a);
#endif

// POV-Ray export type
const int TRIANGLE=0;
const int SMOOTH_TRIANGLE=1;
const int TEMP_EXPORT=2; // actually SMOOTH_TRIANGLE and render 

// function Load type {open,import}
const int OPEN=0;
const int IMPORT=1;

const double defOGLAngle=46.826450; // 60 POVAngle in 4:3 win

// right,left,top,bottom
// setting for Windows98,qvwm default
// TODO for window manager
const int glOffsetR=6;
const int glOffsetL=6; 
const int glOffsetT=25;
const int glOffsetB=6; 

// AddViewIndex offset
const int deleteIndexOffset=6;
const int toggleIndexOffset=4;

static char tmpExpFilename[1024];

void AddEdge(int,int);
int FindEdge(const int,const int);
void AddFace(int,int,int,int,int,int);
void AddFace(int,int,int,int,int,int,int,int);
void PreSub(); // pre subdivide
void Subdivide(int); // subdivision times
void SubdivideCore(); // subdivision algorithm
void PostSub(); // post subdivide
void SelectedSmooth();
void SelectedSharp();
void RecalcSub(int);
void WeightPropagate();
void DebugObjects();
void PrintHelp();
void PrintVertex();
void PrintEdge();
void PrintFace();
void PrintVec(vector<int>);
void RemoveExtraFace();
void RemoveSelected();
void Find4Face(int,int,int);
void Sort3(int&,int&,int&);
void RemoveFace(int);
void RemoveEdge(int);
void RemoveVertex(int);
void Extrude();
void ExtrudeCore();
void FourToThree();

// Select
void SelectAll();
void DeselectAll();
void DeselectAllofAll();

// Callback
int CheckSave();
void NewCB();
void OpenCB();
void ChangeCB();
void SetChanged(int);
void SaveCB();
void SaveAsCB();
void QuitCB();

// File
void PYGInit();
void Load(char*,int);
void Save(char*);
void Load();
void Import();

void ExportOFF();
void ExportPOV();
void ExportPOVSmooth();
int GetString(char*,char*);
void ExportPOVCore(int,int);
void ImportPYG();

void Render();

void CreateNewWin();
void ChangeWinProp();

void CreateView(int);
  void CreateCamera(int);
  void CreateGLWUI();
void DeleteView(int);
void PreDeleteView(Fl_Menu_Item*,int);
void AlignViews();

void ToggleView(Fl_Menu_Item*,int);

void CheckBBSCore(Vector3d&);
void CheckBBS(); // check bounding box of selection
void CenteringObject();

double POVAng2OGLAng(Camera&,double);
double OGLAng2POVAng(Camera&,double);

void PushOP(int);
void Undo();
void Redo();

template<typename T>
void VectorCopy(vector<T>&,vector<T>&,int,int);

void Unique(vector<int>&);
void RefineTotalV();
#endif // _PYGMALION_H_
