#include <iostream>
#include <math.h>
#include "vector3d.h"

using namespace std; 

// Constructor
Vector3d::Vector3d(){
    coord[0]=0;
    coord[1]=0;
    coord[2]=0;
}

// Constructor
Vector3d::Vector3d(const double& d1,const double& d2,const double& d3){
    coord[0]=d1;
    coord[1]=d2;
    coord[2]=d3;
}

// Copy Constructor
Vector3d::Vector3d(const Vector3d& rhs){
  coord[0]=rhs.coord[0];
  coord[1]=rhs.coord[1];
  coord[2]=rhs.coord[2];
}

// Destructor
Vector3d::~Vector3d(){
}

// Substitution operator
const Vector3d&
Vector3d::operator=(const Vector3d& rhs){
    if (this==&rhs)
    return *this;

    coord[0]=rhs.coord[0];
    coord[1]=rhs.coord[1];
    coord[2]=rhs.coord[2];
    return *this;
}

// Plus Operator
Vector3d
Vector3d::operator+(const Vector3d& rhs){
    Vector3d tmp;
    tmp.coord[0]=coord[0]+rhs.coord[0];
    tmp.coord[1]=coord[1]+rhs.coord[1];
    tmp.coord[2]=coord[2]+rhs.coord[2];
    return tmp;
}

// Minus Operator
Vector3d
Vector3d::operator-(const Vector3d& rhs){
    Vector3d tmp;
    tmp.coord[0]=coord[0]-rhs.coord[0];
    tmp.coord[1]=coord[1]-rhs.coord[1];
    tmp.coord[2]=coord[2]-rhs.coord[2];
    return tmp;
}

// Multi Operator
Vector3d
Vector3d::operator*(const double& rhs){
    Vector3d tmp;
    tmp.coord[0]=coord[0]*rhs;
    tmp.coord[1]=coord[1]*rhs;
    tmp.coord[2]=coord[2]*rhs;
    return tmp;
}

/*
ostream
Vector3d::operator<<(Vector3d& out){
  ostream value;
  value << out.coord[0] << out.coord[1] << out.coord[2];
  return value;
}
*/

// Plus Vector3d Value
void
Vector3d::Plus(const Vector3d& src){
    coord[0]+=src.coord[0];
    coord[1]+=src.coord[1];
    coord[2]+=src.coord[2];
}

// Minus Vector3d Value
void
Vector3d::Minus(const Vector3d& src){
    coord[0]-=src.coord[0];
    coord[1]-=src.coord[1];
    coord[2]-=src.coord[2];
}

// Multiply 
void
Vector3d::Multi(const double& mul){
    coord[0]*=mul;
    coord[1]*=mul;
    coord[2]*=mul;    
}

// Cross Product
const Vector3d
Vector3d::Cross(const Vector3d& src){
    Vector3d tmp;
    tmp.coord[0]=coord[1]*src.coord[2] - coord[2]*src.coord[1];
    tmp.coord[1]=coord[2]*src.coord[0] - coord[0]*src.coord[2];
    tmp.coord[2]=coord[0]*src.coord[1] - coord[1]*src.coord[0];
    return tmp;
}

// Cross Product 2
void
Vector3d::Cross(const Vector3d& src2,const Vector3d& src1){
  coord[0]=src2.coord[1]*src1.coord[2] - src2.coord[2]*src1.coord[1];
  coord[1]=src2.coord[2]*src1.coord[0] - src2.coord[0]*src1.coord[2];
  coord[2]=src2.coord[0]*src1.coord[1] - src2.coord[1]*src1.coord[0];
}


// Dot Product
const double 
Vector3d::Dot(const Vector3d& src){
    double tmp;
    tmp=coord[0]*src.coord[0];
    tmp+=coord[1]*src.coord[1];
    tmp+=coord[2]*src.coord[2];
    return tmp;
}

// Normalize
void
Vector3d::Normalize(){
  double tmp;
  
  tmp=sqrt(this->Dot(*this));
  this->Multi(1.0/tmp);  
}

// Multi Matrix[3][3]
void
Vector3d::MultiMat33(double* mat){
  double coord0=coord[0];
  double coord1=coord[1];
  double coord2=coord[2];
  coord[0]=mat[0]*coord0+mat[1]*coord1+mat[2]*coord2;
  coord[1]=mat[3]*coord0+mat[4]*coord1+mat[5]*coord2;
  coord[2]=mat[6]*coord0+mat[7]*coord1+mat[8]*coord2;
}

// Set Coordinates
void
Vector3d::Set(const double& xx,const double& yy,const double& zz){
    coord[0]=xx;
    coord[1]=yy;
    coord[2]=zz;
}

// Get Coordinates
void
Vector3d::Get(double& xx,double& yy,double& zz){
    xx=coord[0];
    yy=coord[1];
    zz=coord[2];
}

// Print Coordinates
void
Vector3d::Print(){
    cout << coord[0] << "," << coord[1] << "," << coord[2] << endl; 
}

// Return X
double
Vector3d::X(){
    return coord[0];
}

// Return Y
double
Vector3d::Y(){
    return coord[1];
}

// Return Z
double
Vector3d::Z(){
    return coord[2];
}

void
Vector3d::WriteFile(FILE* fp){
  fprintf(fp,"%f %f %f\n",coord[0],coord[1],coord[2]);
}
